##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Multistage steam turbine for power generation.

Liese, (2014). "Modeling of a Steam Turbine Including Partial Arc Admission
    for Use in a Process Simulation Software Environment." Journal of Engineering
    for Gas Turbines and Power. v136, November
"""
import copy

from pyomo.environ import (
    RangeSet,
    Set,
    TransformationFactory,
    Var,
    value,
    SolverFactory,
)
from pyomo.network import Arc
from pyomo.common.config import ConfigBlock, ConfigValue, In

from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    EnergyBalanceType,
    MomentumBalanceType,
    MaterialBalanceType,
    useDefault,
)
from idaes.unit_models import (
    Separator,
    Mixer,
    SplittingType,
    EnergySplittingType,
    MomentumMixingType,
)
from idaes.unit_models.power_generation import (
    TurbineInletStage,
    TurbineStage,
    TurbineOutletStage,
    SteamValve,
)
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util import from_json, to_json, StoreSpec
from idaes.core.util.misc import copy_port_values as _set_port
from pyomo.common.config import ConfigBlock, ConfigValue, In, ConfigList
from idaes.core.util.config import is_physical_parameter_block
from idaes.logger import getIdaesLogger, getInitLogger, init_tee, condition

_log = getIdaesLogger(__name__)


def _define_turbine_multistage_config(config):
    config.declare(
        "dynamic",
        ConfigValue(
            domain=In([True, False]),
            default=False,
            description="Dynamic model flag",
            doc="Indicates whether the model is dynamic.",
        ),
    )
    config.declare(
        "has_holdup",
        ConfigValue(
            default=useDefault,
            domain=In([useDefault, True, False]),
            description="Holdup construction flag",
            doc="""Indicates whether holdup terms should be constructed or not.
Must be True if dynamic = True,
**default** - False.
**Valid values:** {
**True** - construct holdup terms,
**False** - do not construct holdup terms}""",
        ),
    )
    config.declare(
        "has_phase_equilibrium",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Calculate phase equilibrium in mixed stream",
            doc="""Argument indicating whether phase equilibrium should be
calculated for the resulting mixed stream,
**default** - False.
**Valid values:** {
**True** - calculate phase equilibrium in mixed stream,
**False** - do not calculate equilibrium in mixed stream.}""",
        ),
    )
    config.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.componentTotal,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.componentTotal`.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    config.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    config.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    config.declare(
        "num_parallel_inlet_stages",
        ConfigValue(
            default=4,
            domain=int,
            description="Number of parallel inlet stages to simulate partial arc "
            "admission.  Default=4",
        ),
    )
    config.declare(
        "num_hp",
        ConfigValue(
            default=2,
            domain=int,
            description="Number of high pressure stages not including inlet stage",
            doc="Number of high pressure stages not including inlet stage",
        ),
    )
    config.declare(
        "num_ip",
        ConfigValue(
            default=10,
            domain=int,
            description="Number of intermediate pressure stages",
            doc="Number of intermediate pressure stages",
        ),
    )
    config.declare(
        "num_lp",
        ConfigValue(
            default=5,
            domain=int,
            description="Number of low pressure stages not including outlet stage",
            doc="Number of low pressure stages not including outlet stage",
        ),
    )
    config.declare(
        "hp_split_locations",
        ConfigList(
            default=[],
            domain=int,
            description="Locations of splitters in HP section",
            doc="A list of index locations of splitters in the HP section. The "
            "indexes indicate after which stage to include splitters.  0 is "
            "between the inlet stage and the first regular HP stage.",
        ),
    )
    config.declare(
        "ip_split_locations",
        ConfigList(
            default=[],
            domain=int,
            description="Locations of splitters in IP section",
            doc="A list of index locations of splitters in the IP section. The "
            "indexes indicate after which stage to include splitters.",
        ),
    )
    config.declare(
        "lp_split_locations",
        ConfigList(
            default=[],
            domain=int,
            description="Locations of splitter in LP section",
            doc="A list of index locations of splitters in the LP section. The "
            "indexes indicate after which stage to include splitters.",
        ),
    )
    config.declare(
        "hp_disconnect",
        ConfigList(
            default=[],
            domain=int,
            description="HP Turbine stages to not connect to next with an arc.",
            doc="HP Turbine stages to not connect to next with an arc. This is "
            "usually used to insert addtional units between stages on a "
            "flowsheet, such as a reheater",
        ),
    )
    config.declare(
        "ip_disconnect",
        ConfigList(
            default=[],
            domain=int,
            description="IP Turbine stages to not connect to next with an arc.",
            doc="IP Turbine stages to not connect to next with an arc. This is "
            "usually used to insert addtional units between stages on a "
            "flowsheet, such as a reheater",
        ),
    )
    config.declare(
        "lp_disconnect",
        ConfigList(
            default=[],
            domain=int,
            description="LP Turbine stages to not connect to next with an arc.",
            doc="LP Turbine stages to not connect to next with an arc. This is "
            "usually used to insert addtional units between stages on a "
            "flowsheet, such as a reheater",
        ),
    )
    config.declare(
        "hp_split_num_outlets",
        ConfigValue(
            default={},
            domain=dict,
            description="Dict, hp split index: number of splitter outlets, if not 2",
        ),
    )
    config.declare(
        "ip_split_num_outlets",
        ConfigValue(
            default={},
            domain=dict,
            description="Dict, ip split index: number of splitter outlets, if not 2",
        ),
    )
    config.declare(
        "lp_split_num_outlets",
        ConfigValue(
            default={},
            domain=dict,
            description="Dict, lp split index: number of splitter outlets, if not 2",
        ),
    )


@declare_process_block_class(
    "TurbineMultistage",
    doc="Multistage steam turbine with optional reheat and extraction",
)
class TurbineMultistageData(UnitModelBlockData):

    CONFIG = ConfigBlock()
    _define_turbine_multistage_config(CONFIG)

    def build(self):
        super(TurbineMultistageData, self).build()
        config = self.config
        unit_cfg = {  # general unit model config
            "dynamic": config.dynamic,
            "has_holdup": config.has_holdup,
            "has_phase_equilibrium": config.has_phase_equilibrium,
            "material_balance_type": config.material_balance_type,
            "property_package": config.property_package,
            "property_package_args": config.property_package_args,
        }
        ni = self.config.num_parallel_inlet_stages
        inlet_idx = self.inlet_stage_idx = RangeSet(ni)

        # Adding unit models
        # ------------------------

        # Splitter to inlet that splits main flow into parallel flows for
        # paritial arc admission to the turbine
        self.inlet_split = Separator(default=self._split_cfg(unit_cfg, ni))
        self.throttle_valve = SteamValve(inlet_idx, default=unit_cfg)
        self.inlet_stage = TurbineInletStage(inlet_idx, default=unit_cfg)
        # mixer to combine the parallel flows back together
        self.inlet_mix = Mixer(default=self._mix_cfg(unit_cfg, ni))
        # add turbine sections.
        # inlet stage -> hp stages -> ip stages -> lp stages -> outlet stage
        self.hp_stages = TurbineStage(RangeSet(config.num_hp), default=unit_cfg)
        self.ip_stages = TurbineStage(RangeSet(config.num_ip), default=unit_cfg)
        self.lp_stages = TurbineStage(RangeSet(config.num_lp), default=unit_cfg)
        self.outlet_stage = TurbineOutletStage(default=unit_cfg)

        for i in self.hp_stages:
            self.hp_stages[i].ratioP.fix()
            self.hp_stages[i].efficiency_isentropic[:].fix()
        for i in self.ip_stages:
            self.ip_stages[i].ratioP.fix()
            self.ip_stages[i].efficiency_isentropic[:].fix()
        for i in self.lp_stages:
            self.lp_stages[i].ratioP.fix()
            self.lp_stages[i].efficiency_isentropic[:].fix()

        # Then make splitter config.  If number of outlets is specified
        # make a specific config, otherwise use default with 2 outlets
        s_sfg_default = self._split_cfg(unit_cfg, 2)
        hp_splt_cfg = {}
        ip_splt_cfg = {}
        lp_splt_cfg = {}
        # Now to finish up if there are more than two outlets, set that
        for i, v in config.hp_split_num_outlets.items():
            hp_splt_cfg[i] = self._split_cfg(unit_cfg, v)
        for i, v in config.ip_split_num_outlets.items():
            ip_splt_cfg[i] = self._split_cfg(unit_cfg, v)
        for i, v in config.lp_split_num_outlets.items():
            lp_splt_cfg[i] = self._split_cfg(unit_cfg, v)
        # put in splitters for turbine steam extractions
        if config.hp_split_locations:
            self.hp_split = Separator(
                config.hp_split_locations, default=s_sfg_default, initialize=hp_splt_cfg
            )
        if config.ip_split_locations:
            self.ip_split = Separator(
                config.ip_split_locations, default=s_sfg_default, initialize=ip_splt_cfg
            )
        if config.lp_split_locations:
            self.lp_split = Separator(
                config.lp_split_locations, default=s_sfg_default, initialize=lp_splt_cfg
            )

        # Done with unit models.  Adding Arcs (streams).
        # ------------------------------------------------

        # First up add streams in the inlet section
        def _split_to_rule(b, i):
            return {
                "source": getattr(self.inlet_split, "outlet_{}".format(i)),
                "destination": self.throttle_valve[i].inlet,
            }

        def _valve_to_rule(b, i):
            return {
                "source": self.throttle_valve[i].outlet,
                "destination": self.inlet_stage[i].inlet,
            }

        def _inlet_to_rule(b, i):
            return {
                "source": self.inlet_stage[i].outlet,
                "destination": getattr(self.inlet_mix, "inlet_{}".format(i)),
            }

        self.split_to_valve_stream = Arc(inlet_idx, rule=_split_to_rule)
        self.valve_to_inlet_stage_stream = Arc(inlet_idx, rule=_valve_to_rule)
        self.inlet_stage_to_mix = Arc(inlet_idx, rule=_inlet_to_rule)

        # There are three sections HP, IP, and LP which all have the same sort
        # of internal connctions, so the functions below provide some generic
        # capcbilities for adding the internal Arcs (streams).
        def _arc_indexes(nstages, index_set, discon, splits):
            """
            This takes the index set of all possible streams in a turbine
            section and throws out arc indexes for stages that are disconnected
            and arc indexes that are not needed because there is no splitter
            after a stage.

            Args:
                nstages (int): Number of stages in section
                index_set (Set): Index set for arcs in the section
                discon (list): Disconnected stages in the section
                splits (list): Spliter locations
            """
            sr = set()  # set of things to remove from the Arc index set
            for i in index_set:
                if (i[0] in discon or i[0] == nstages) and i[0] in splits:
                    # don't connect stage i to next remove stream after split
                    sr.add((i[0], 2))
                elif (i[0] in discon or i[0] == nstages) and i[0] not in splits:
                    # no splitter and disconnect so remove both streams
                    sr.add((i[0], 1))
                    sr.add((i[0], 2))
                elif i[0] not in splits:
                    # no splitter and not disconnected so just second stream
                    sr.add((i[0], 2))
                else:
                    # has splitter so need both streams don't remove anything
                    pass
            for i in sr:  # remove the unneeded Arc indexes
                index_set.remove(i)

        def _arc_rule(turbines, splitters):
            """
            This creates a rule function for arcs in a turbine section. When
            this is used the indexes for nonexistant stream will have already
            been removed, so any indexes the rule will get should have a stream
            associated.

            Args:
                turbines (TurbineStage): Indexed block with turbine section stages
                splitters (Separator): Indexed block of splitters
            """

            def _rule(b, i, j):
                if i in splitters and j == 1:
                    return {
                        "source": turbines[i].outlet,
                        "destination": splitters[i].inlet,
                    }
                elif j == 2:
                    return {
                        "source": splitters[i].outlet_1,
                        "destination": turbines[i + 1].inlet,
                    }
                else:
                    return {
                        "source": turbines[i].outlet,
                        "destination": turbines[i + 1].inlet,
                    }

            return _rule

        # Create initial arcs index sets with all possible streams
        self.hp_stream_idx = Set(initialize=self.hp_stages.index_set() * [1, 2])
        self.ip_stream_idx = Set(initialize=self.ip_stages.index_set() * [1, 2])
        self.lp_stream_idx = Set(initialize=self.lp_stages.index_set() * [1, 2])

        # Throw out unneeded streams
        _arc_indexes(
            config.num_hp,
            self.hp_stream_idx,
            config.hp_disconnect,
            config.hp_split_locations,
        )
        _arc_indexes(
            config.num_ip,
            self.ip_stream_idx,
            config.ip_disconnect,
            config.ip_split_locations,
        )
        _arc_indexes(
            config.num_lp,
            self.lp_stream_idx,
            config.lp_disconnect,
            config.lp_split_locations,
        )

        # Create connections internal to each turbine section (hp, ip, and lp)
        self.hp_stream = Arc(
            self.hp_stream_idx, rule=_arc_rule(self.hp_stages, self.hp_split)
        )
        self.ip_stream = Arc(
            self.ip_stream_idx, rule=_arc_rule(self.ip_stages, self.ip_split)
        )
        self.lp_stream = Arc(
            self.lp_stream_idx, rule=_arc_rule(self.lp_stages, self.lp_split)
        )

        # Connect hp section to ip section unless its a disconnect location
        last_hp = config.num_hp
        if 0 not in config.ip_disconnect and last_hp not in config.hp_disconnect:
            if last_hp in config.hp_split_locations:  # connect splitter to ip
                self.hp_to_ip_stream = Arc(
                    source=self.hp_split[last_hp].outlet_1,
                    destination=self.ip_stages[1].inlet,
                )
            else:  # connect last hp to ip
                self.hp_to_ip_stream = Arc(
                    source=self.hp_stages[last_hp].outlet,
                    destination=self.ip_stages[1].inlet,
                )
        # Connect ip section to lp section unless its a disconnect location
        last_ip = config.num_ip
        if 0 not in config.lp_disconnect and last_ip not in config.ip_disconnect:
            if last_ip in config.ip_split_locations:  # connect splitter to ip
                self.ip_to_lp_stream = Arc(
                    source=self.ip_split[last_ip].outlet_1,
                    destination=self.lp_stages[1].inlet,
                )
            else:  # connect last hp to ip
                self.ip_to_lp_stream = Arc(
                    source=self.ip_stages[last_ip].outlet,
                    destination=self.lp_stages[1].inlet,
                )
        # Connect inlet stage to hp section
        #   not allowing disconnection of inlet and first regular hp stage
        if 0 in config.hp_split_locations:
            # connect inlet mix to splitter and splitter to hp section
            self.inlet_to_splitter_stream = Arc(
                source=self.inlet_mix.outlet, destination=self.hp_split[0].inlet
            )
            self.splitter_to_hp_stream = Arc(
                source=self.hp_split[0].outlet_1, destination=self.hp_stages[1].inlet
            )
        else:  # connect mixer to first hp turbine stage
            self.inlet_to_hp_stream = Arc(
                source=self.inlet_mix.outlet, destination=self.hp_stages[1].inlet
            )

        @self.Expression(self.flowsheet().config.time)
        def power(b, t):
            return (
                sum(b.inlet_stage[i].power_shaft[t] for i in b.inlet_stage)
                + b.outlet_stage.power_shaft[t]
                + sum(b.hp_stages[i].power_shaft[t] for i in b.hp_stages)
                + sum(b.ip_stages[i].power_shaft[t] for i in b.ip_stages)
                + sum(b.lp_stages[i].power_shaft[t] for i in b.lp_stages)
            )

        # Connect inlet stage to hp section
        #   not allowing disconnection of inlet and first regular hp stage
        last_lp = config.num_lp
        if last_lp in config.lp_split_locations:  # connect splitter to outlet
            self.lp_to_outlet_stream = Arc(
                source=self.lp_split[last_lp].outlet_1,
                destination=self.outlet_stage.inlet,
            )
        else:  # connect last lpstage to outlet
            self.lp_to_outlet_stream = Arc(
                source=self.lp_stages[last_lp].outlet,
                destination=self.outlet_stage.inlet,
            )
        TransformationFactory("network.expand_arcs").apply_to(self)

    def _split_cfg(self, unit_cfg, no=2):
        """
        This creates a configuration dictionary for a splitter.

        Args:
            unit_cfg: The base unit config dict.
            no: Number of outlets, default=2
        """
        # Create a dict for splitter config args
        s_cfg = copy.copy(unit_cfg)  # splitter config based on unit_cfg
        s_cfg.update(
            split_basis=SplittingType.totalFlow,
            ideal_separation=False,
            num_outlets=no,
            energy_split_basis=EnergySplittingType.equal_molar_enthalpy,
        )
        return s_cfg

    def _mix_cfg(self, unit_cfg, ni=2):
        """
        This creates a configuration dictionary for a mixer.

        Args:
            unit_cfg: The base unit config dict.
            ni: Number of inlets, default=2
        """
        m_cfg = copy.copy(unit_cfg)  # splitter config based on unit_cfg
        m_cfg.update(
            num_inlets=ni, momentum_mixing_type=MomentumMixingType.minimize_and_equality
        )
        return m_cfg

    def throttle_cv_fix(self, value):
        """
        Fix the thottle valve coefficients.  These are generally the same for
        each of the parallel stages so this provides a convenient way to set
        them.

        Args:
            value: The value to fix the turbine inlet flow coefficients at
        """
        for i in self.throttle_valve:
            self.throttle_valve[i].Cv.fix(value)

    def turbine_inlet_cf_fix(self, value):
        """
        Fix the inlet turbine stage flow coefficient.  These are
        generally the same for each of the parallel stages so this provides
        a convenient way to set them.

        Args:
            value: The value to fix the turbine inlet flow coefficients at
        """
        for i in self.inlet_stage:
            self.inlet_stage[i].flow_coeff.fix(value)

    def turbine_outlet_cf_fix(self, value):
        """
        Fix the inlet turbine stage flow coefficient.  These are
        generally the same for each of the parallel stages so this provides
        a convenient way to set them.

        Args:
            value: The value to fix the turbine inlet flow coefficients at
        """
        self.outlet_stage.flow_coeff.fix(value)

    def initialize(
        self, outlvl=6, solver="ipopt", optarg={"tol": 1e-6, "max_iter": 35}
    ):
        """
        Initialize
        """
        # sp is what to save to make sure state after init is same as the start
        #   saves value, fixed, and active state, doesn't load originally free
        #   values, this makes sure original problem spec is same but initializes
        #   the values of free vars
        init_log = getInitLogger(self.name, outlvl)

        sp = StoreSpec.value_isfixed_isactive(only_fixed=True)
        istate = to_json(self, return_dict=True, wts=sp)
        ni = self.config.num_parallel_inlet_stages
        flow_guess = self.inlet_split.inlet.flow_mol[0].value

        def init_section(stages, splits, disconnects, prev_port):
            if 0 in splits:
                _set_port(splits[0].inlet, prev_port)
                splits[0].initialize(outlvl=outlvl, solver=solver, optarg=optarg)
                prev_port = splits[0].outlet_1
            for i in stages:
                if i - 1 not in disconnects:
                    _set_port(stages[i].inlet, prev_port)
                stages[i].initialize(outlvl=outlvl, solver=solver, optarg=optarg)
                prev_port = stages[i].outlet
                if i in splits:
                    _set_port(splits[i].inlet, prev_port)
                    splits[i].initialize(outlvl=outlvl, solver=solver, optarg=optarg)
                    prev_port = splits[i].outlet_1
            return prev_port

        for k in [1]:
            # Initialize Splitter
            # Fix n - 1 split fractions
            self.inlet_split.split_fraction[0, "outlet_1"].value = 1.0 / ni
            for i in self.inlet_stage_idx:
                if i == 1:  # fix rest of splits at leaving first one free
                    continue
                self.inlet_split.split_fraction[0, "outlet_{}".format(i)].fix(1.0 / ni)
            # fix inlet and free outlet
            self.inlet_split.inlet.fix()
            for i in self.inlet_stage_idx:
                ol = getattr(self.inlet_split, "outlet_{}".format(i))
                ol.unfix()
            self.inlet_split.initialize(outlvl=outlvl, solver=solver, optarg=optarg)
            # free split fractions
            for i in self.inlet_stage_idx:
                self.inlet_split.split_fraction[0, "outlet_{}".format(i)].unfix()

            # Initialize valves
            for i in self.inlet_stage_idx:
                _set_port(
                    self.throttle_valve[i].inlet,
                    getattr(self.inlet_split, "outlet_{}".format(i)),
                )
                self.throttle_valve[i].initialize(
                    outlvl=outlvl, solver=solver, optarg=optarg
                )

            # Initialize turbine
            for i in self.inlet_stage_idx:
                _set_port(self.inlet_stage[i].inlet, self.throttle_valve[i].outlet)
                self.inlet_stage[i].initialize(
                    outlvl=outlvl, solver=solver, optarg=optarg
                )

            # Initialize Mixer
            self.inlet_mix.use_minimum_inlet_pressure_constraint()
            for i in self.inlet_stage_idx:
                _set_port(
                    getattr(self.inlet_mix, "inlet_{}".format(i)),
                    self.inlet_stage[i].outlet,
                )
                getattr(self.inlet_mix, "inlet_{}".format(i)).fix()
            self.inlet_mix.initialize(outlvl=outlvl, solver=solver, optarg=optarg)
            for i in self.inlet_stage_idx:
                getattr(self.inlet_mix, "inlet_{}".format(i)).unfix()
            self.inlet_mix.use_equal_pressure_constraint()

            prev_port = self.inlet_mix.outlet
            prev_port = init_section(
                self.hp_stages, self.hp_split, self.config.hp_disconnect, prev_port
            )
            if len(self.hp_stages) in self.config.hp_disconnect:
                prev_port = self.ip_stages[1].inlet
            prev_port = init_section(
                self.ip_stages, self.ip_split, self.config.ip_disconnect, prev_port
            )
            if len(self.ip_stages) in self.config.ip_disconnect:
                prev_port = self.lp_stages[1].inlet
            prev_port = init_section(
                self.lp_stages, self.lp_split, self.config.lp_disconnect, prev_port
            )

            _set_port(self.outlet_stage.inlet, prev_port)
            self.outlet_stage.initialize(outlvl=outlvl, solver=solver, optarg=optarg)
            for t in self.flowsheet().time:
                self.inlet_split.inlet.flow_mol[
                    t
                ].value = self.outlet_stage.inlet.flow_mol[t].value

        slvr = SolverFactory(solver)
        slvr.options = optarg
        init_log.log(4, "Solve full multistage turbine")
        self.inlet_split.inlet.flow_mol.unfix()
        res = slvr.solve(self, tee=init_tee(init_log))

        init_log.log(
            3,
            "Flow guess: {}, Initialized Flow: {}".format(
                flow_guess, self.outlet_stage.inlet.flow_mol[0].value
            ),
        )
        init_log.log(5, "Initialization Complete: {}".format(condition(res)))

        from_json(self, sd=istate, wts=sp)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import numpy as np
noise = 0.1


def sim(data):
    x0 = data
    params = [1.5,2.1,0.9]
    #pyomo simulator for cracking example
    import numpy as np
    import pyomo.environ as pyo

    dshape = np.shape(x0)
    
    if len(dshape) == 1:
        x0 = np.expand_dims(x0, axis=-1)
        x0 = np.ndarray.transpose(x0)
        dshape = np.shape(x0)
    npc = dshape[0]
    ns = dshape[1]
    if dshape[0] == 1:
        xo = [x0]*npc
    
    def pyomosim(x0):
        # Define rate parameters 
        flow = 1.0
        vol = 1.0
        gc = 8.314
        # Define kinetic rate parameters
        k = params        
        a0,b0,c0,d0,e0 = np.array(x0).T.tolist()
        opt = pyo.SolverFactory('baron')
        model = pyo.ConcreteModel()
#        bound_ub = 100
        # Define model variables
        model.a = pyo.Var(domain = pyo.NonNegativeReals)#, bounds = (0.0,bound_ub))
        model.b = pyo.Var(domain = pyo.NonNegativeReals)#, bounds = (0.0,bound_ub))
        model.c = pyo.Var(domain = pyo.NonNegativeReals)#, bounds = (0.0,bound_ub))
        model.d = pyo.Var(domain = pyo.NonNegativeReals)#, bounds = (0.0,bound_ub))
        model.e = pyo.Var(domain = pyo.NonNegativeReals)#, bounds = (0.0,bound_ub))
        model.dset = pyo.RangeSet(5)
        model.dum = pyo.Var(model.dset)

        model.r1 = pyo.Var(domain = pyo.Reals)
        model.r2 = pyo.Var(domain = pyo.Reals)
        model.r3 = pyo.Var(domain = pyo.Reals)
        
        def fr1(model):
            return model.r1 == k[0] * model.a * model.b
        def fr2(model):
            return model.r2 == k[1] * model.b * model.c
        def fr3(model):
            return model.r3 == k[2] * model.a * model.d

        model.er1 = pyo.Constraint( rule = fr1)
        model.er2 = pyo.Constraint( rule = fr2)
        model.er3 = pyo.Constraint( rule = fr3)
        
        num = 1.0
        def fra(model):
            return num * model.dum[1] == (flow/vol)*(a0-model.a ) - model.r1 -model.r3
        def frb(model):
            return num * model.dum[2] == (flow/vol)*(b0-model.b ) - model.r1 - model.r2
        def frc(model):
            return num * model.dum[3] == (flow/vol)*(c0-model.c) + model.r1 - model.r2
        def frd(model):
            return num * model.dum[4] == (flow/vol)*(d0-model.d) + model.r2 - model.r3
        def fre(model):
            return num * model.dum[5] == (flow/vol)*(e0-model.e) + model.r3

        model.era = pyo.Constraint(rule=fra)
        model.erb = pyo.Constraint( rule = frb)
        model.erc = pyo.Constraint( rule = frc)
        model.erd = pyo.Constraint( rule = frd)
        model.ere = pyo.Constraint( rule = fre)

        def objf(model):
            return sum([ model.dum[i]**2 for i in model.dset])
        
        model.OBJ = pyo.Objective(rule = objf)

        results = opt.solve(model, tee=False)
        model.solutions.store_to(results)
        # Note: noise is not truly normally distributed as concentration values cannot be negative
#        print 'debug this : ', results.Solution.Variable['a']['Value'],np.random.normal(0,noise*results.Solution.Variable['a']['Value']),np.argmax([0.0,results.Solution.Variable['a']['Value']+np.random.normal(0,noise*results.Solution.Variable['a']['Value'])])
        v = [results.Solution.Variable[key]['Value'] for key in ['a','b','c','d','e']]
        vn = [results.Solution.Variable[key]['Value']+np.random.normal(0,noise*results.Solution.Variable[key]['Value']) for key in ['a','b','c','d','e']]
        tsum = 0
        for i in range(5):
            if vn[i] < 0:
                vn[i] = v[i]
                tsum+= 1
#        print 'total number of zeros : ', tsum
        return vn
    
    # Simulate over requested datapoints
    concentrations = []
    for i in range(npc):
        conres = pyomosim(x0[i])
        concentrations.append(conres)
    return concentrations



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Exception classes.
"""
import logging

__author__ = 'Dan Gunter <dkgunter@lbl.gov>'

_log = logging.getLogger(__name__)


class DMFError(Exception):
    def __init__(self, detailed_error='No details'):
        msg = '{}'.format(detailed_error)
        super(DMFError, self).__init__(msg)


class ParseError(Exception):
    pass


class CommandError(Exception):
    def __init__(self, command, operation, details):
        msg = 'Operation "{op}" in command "{c}" failed: {d}'.format(
            op=operation, c=command, d=details)
        super(CommandError, self).__init__(msg)


class WorkspaceError(DMFError):
    pass


class WorkspaceNotFoundError(WorkspaceError):
    def __init__(self, from_dir):
        msg = 'Workspace not found for path "{}" '.format(from_dir)
        super(WorkspaceNotFoundError, self).__init__(msg)


class WorkspaceCannotCreateError(WorkspaceError):
    def __init__(self, path):
        msg = 'Unable to create new workspace at "{}"'.format(path)
        super(WorkspaceCannotCreateError, self).__init__(msg)


class WorkspaceConfNotFoundError(WorkspaceError):
    def __init__(self, path):
        msg = 'Workspace config not found at path "{}" '.format(path)
        super(WorkspaceConfNotFoundError, self).__init__(msg)


class WorkspaceConfMissingField(WorkspaceError):
    def __init__(self, path, name, desc):
        msg = 'Workspace config at path "{}" missing {} field "{}"'\
            .format(path, desc, name)
        super(WorkspaceConfMissingField, self).__init__(msg)


class FileError(Exception):
    pass


class ResourceError(Exception):
    pass


class NoSuchResourceError(ResourceError):
    def __init__(self, name=None, id_=None):
        if name is not None and id_ is None:
            msg = 'No resource of type "{}" found'.format(name)
        elif id_ is not None and name is None:
            msg = 'No resource "{}" found'.format(id_)
        elif name is None and id_ is None:
            msg = 'Resource not found'
        else:
            msg = 'No resource "{}" of type "{}" found'.format(id_, name)
        super(NoSuchResourceError, self).__init__(msg)


class DuplicateResourceError(ResourceError):
    def __init__(self, op, id_):
        msg = 'While executing "{}": Duplicate resource "{}"'\
            .format(op, id_)
        super(DuplicateResourceError, self).__init__(msg)


class BadResourceError(ResourceError):
    pass


class SearchError(Exception):
    def __init__(self, spec, problem):
        msg = 'Search "{}" failed: {}'.format(spec, problem)
        super(SearchError, self).__init__(msg)


class ModuleFormatError(Exception):
    def __init__(self, module_name, type_, what):
        msg = 'Python module "{}" does not conform to conventions for a '\
              '{} module: {}'.format(module_name, type_, what)
        super(ModuleFormatError, self).__init__(msg)


class DmfError(Exception):
    pass


class InvalidRelationError(DmfError):
    def __init__(self, subj, pred, obj):
        msg = 'Invalid relation: {} --({})--> {}'.format(subj, pred, obj)
        super(InvalidRelationError, self).__init__(msg)


class DataFormatError(DmfError):
    def __init__(self, dtype, err):
        msg = 'Bad data format for type "{}":\n{}'.format(dtype, err)
        super(DataFormatError, self).__init__(msg)

# Alamo


class AlamoError(DmfError):
    def __init__(self, msg):
        super(AlamoError, self).__init__('ALAMO Error: {}'.format(msg))


class AlamoDisabledError(AlamoError):
    def __init__(self):
        super(AlamoDisabledError, self).__init__('ALAMO is disabled')



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Perform all logic, input, output of commands that is
particular to the CLI.

Call functions defined in 'api' module to handle logic
that is common to the API and CLI.
"""
# stdlib
import glob
import json
import logging
import math
import os
import re
import sys

# Third-party
import jsonschema
import pendulum

# Local
from .dmfbase import DMF, DMFConfig
from .util import strlist
from .util import is_jupyter_notebook, is_python, is_resource_json
from .util import ColorTerm
from .errors import (
    ParseError,
    CommandError,
    WorkspaceNotFoundError,
    WorkspaceConfNotFoundError,
    WorkspaceError,
    BadResourceError,
)
from . import resource
from .workspace import Workspace, find_workspaces

__author__ = "Dan Gunter"

_log = logging.getLogger(__name__)


def workspace_init(dirname, metadata):
    # type: (str, dict) -> None
    """Initialize from root at `dirname`, set environment variable
    for other commands, and parse config file.
    """
    try:
        ws = Workspace(dirname, create=True, add_defaults=True)
    except OSError as err:
        raise CommandError("init", "initialize workspace", str(err))
    except ParseError as err:
        raise CommandError("init", "parse config", str(err))
    except WorkspaceError as err:
        raise CommandError("init", "initialize workspace", str(err))
    _log.info("Created new workspace in: {}".format(dirname))
    if metadata:
        ws.set_meta(metadata)
        _log.info("Set metadata for: {}".format(strlist(list(metadata))))


def workspace_info(dirname):
    # type: (str) -> None
    t = ColorTerm()
    try:
        ws = Workspace(dirname, create=False)
    except WorkspaceNotFoundError:
        print("Workspace not found at path: {}".format(dirname))
        raise CommandError("info", "find workspace", "not found at: {}".format(dirname))
    except WorkspaceConfNotFoundError:
        print("No configuration found for workspace for path: {}".format(dirname))
        raise CommandError(
            "info", "find workspace configuration", "not found at: {}".format(dirname)
        )
    num_obj = DMF(path=ws.root).count()
    bullet = " - "
    print(f"\n{t.blue}Workspace")
    if ws.name and (ws.name != "none"):
        if ws.description and (ws.description != "none"):
            print(f"  {t.blue}[{ws.name}] - {ws.description}")
        else:
            print("  {t.blue}{ws.name} - (no description)")
    elif ws.description and (ws.description != "none"):
        print(f"  {t.blue}(no name) - {ws.description}")
    else:
        print(f"  {t.blue}(no name or description)")
    print("\nGeneral information")
    print(f"{bullet}{t.blue}Location = {ws.root}")
    info = ws.meta.copy()
    if "_id" in info:
        print(f"{bullet}{t.blue}Workspace identifier (_id) = {info['_id']}")
        del info["_id"]
    else:
        print(f"{bullet}{t.blue}Workspace identifier (_id) = unknown")
    if "created" in info:
        print(f"{bullet}{t.blue}Created = {info[ws.CONF_CREATED]}")
    else:
        print(f"{bullet}{t.blue}Created = unknown")
    if "modified" in info:
        print(f"{bullet}{t.blue}Modified = {info[ws.CONF_MODIFIED]}")
    else:
        print(f"{bullet}{t.blue}Modified = unknown")
    print(f"{bullet}{t.blue}Num. resources = {num_obj}")
    print(f"\n{t.magenta}{t.bold}Configuration")
    already_shown = (ws.CONF_MODIFIED, ws.CONF_CREATED, ws.CONF_NAME, ws.CONF_DESC)
    for k in info.keys():
        if k in already_shown:
            continue
        v = info[k]
        print(f"{bullet}{t.blue}{k} = {v}")
    print("")


def init_conf(workspace):
    # type: (str) -> int
    """Initialize the workspace.
    """
    t = ColorTerm()
    # Open/create configuration file
    try:
        conf = DMFConfig()
    except IOError as err:
        print(f"Failed to open global configuration: {err}")
        try:
            open(DMFConfig._filename, "w")
        except IOError:
            print("Failed to create new configuration file")
            return -1
        print("Created new configuration file")
        conf = DMFConfig()
    # If a workspace argument is given, save this value,
    # as the default workspace, in the configuration file
    if workspace:
        fullpath = os.path.abspath(workspace)
        conf.c[conf.WORKSPACE] = fullpath
        conf.save()
    # Print contents of configuration file to standard output
    print(
        f"{t.magenta}{t.bold}DMF global configuration{t.reset} "
        f"<{t.green}{conf._filename}>"
    )
    keys = conf.c.keys()
    if keys:
        for k in sorted(keys):
            print(f" > {t.blue}{k}{t.reset} = {t.bold}{conf.c[k]}]")
    else:
        print(f"{t.blue}(empty)")
    return 0


def workspace_import(path, patterns, exit_on_error):
    # type: (str, list, bool) -> int
    """Import files into workspace.

    Args:
        path (str): Target workspace directory
        patterns (list): List of Unix-style glob for files to import.
                       Files are expected to be resource JSON or a
                       Jupyter Notebook.
        exit_on_error (bool): If False, continue trying to import resources
                              even if one or more fail.

    Returns:
        int: Number of things imported
    Raises:
        BadResourceError, if there is a problem
    """
    d = DMF(path)
    count = 0
    for pattern in patterns:
        for filename in glob.glob(pattern):
            # Skip directories
            if os.path.isdir(filename):
                _log.warning('Not importing directory "{}"'.format(filename))
                continue
            # For Jupyter Notebooks, first create a (temporary)
            # JSON resource from the original data.
            if is_jupyter_notebook(filename):
                try:
                    rsrc = _import_jupyternb(filename)
                except ValueError as e:
                    msg = (
                        "Cannot create resource from Jupyter Notebook "
                        '"{}": {}'.format(filename, e)
                    )
                    if exit_on_error:
                        raise BadResourceError(msg)
                    _log.error(msg)
                    continue
            # For Python files, first create a (temporary)
            # JSON resource from the original data.
            elif is_python(filename):
                try:
                    rsrc = _import_python(filename)
                except ValueError as e:
                    msg = "Cannot create resource from Python file " '"{}": {}'.format(
                        filename, e
                    )
                    if exit_on_error:
                        raise BadResourceError(msg)
                    _log.error(msg)
                    continue
            # JSON resource file
            elif is_resource_json(filename):
                try:
                    rsrc = _import_resource(filename)
                except ValueError as e:
                    msg = 'Bad resource from file "{}": {}'.format(filename, e)
                    if exit_on_error:
                        raise BadResourceError(msg)
                    _log.error(msg)
                    continue
            # Generic file import
            else:
                try:
                    rsrc = _import_file(filename)
                except ValueError as e:
                    msg = "Cannot create resource from file " '"{}": {}'.format(
                        filename, e
                    )
                    if exit_on_error:
                        raise BadResourceError(msg)
                    _log.error(msg)
                    continue
            # Resource in hand. Now add it.
            d.add(rsrc)
            count += 1
        return count


def list_workspaces(root, stream=None):
    """List workspaces found from a given root path.

    Args:
        root: root path
        stream: Output stream (must have .write() method)
    """
    workspaces = find_workspaces(root)
    if stream is None or stream == sys.stdout:
        colors = True
    else:
        colors = False
    t = ColorTerm(enabled=colors)
    if colors:
        output_table = [("Path", "Name")]
    else:
        output_table = [("Path", "Name"), ("----", "----")]
    widths = [4, 4]
    any_good_workspaces = False
    for w in sorted(workspaces):
        try:
            ws = Workspace(w)
            output_table.append((w, ws.name))
            widths = [max(len(w), widths[0]), max(len(ws.name), widths[1])]
            any_good_workspaces = True
        except WorkspaceError:
            pass  # XXX: Should we print a warning?
    if not any_good_workspaces:
        # either no paths, or all paths raised an error
        stream.write("ERROR: No valid workspaces found\n")
    else:
        colfmts = ["{{:{:d}s}}".format(width) for width in widths]
        first_row = True
        for row in output_table:
            for i in (0, 1):
                if colors:
                    if first_row:
                        fmt = f"{t.bold}{colfmts[i]}"
                    else:
                        fmt = f"{[t.blue, t.white][i]}{colfmts[i]}"
                    fmt += t.reset
                else:
                    fmt = colfmts[i]
                stream.write(fmt.format(row[i]))
                stream.write("\n" if i == 1 else " ")
            first_row = False


def list_resources(path, long_format=None, relations=False):
    """List resources in a given DMF workspace.

    Args:
        path (str): Path to the workspace
        long_format (bool): List in long format flag
        relations (bool): Show relationships, in long format

    Returns:
        None
    """
    t = ColorTerm()
    d = DMF(path)
    if long_format:
        resources = list(d.find())
        uuid_pfx = _uuid_prefix([r.uuid for r in resources])
        fields = ("uuid", "name", "type", "modified", "created")
        widths = (uuid_pfx, 30, 20, 19, 19)
        colors = (t.green, t.white, t.yellow, t.white, t.white)
        fmts = [f"{{:{w}s}}" for w in widths]
        left_gutter = "| " if relations else ""
        # table header
        print(
            " " * len(left_gutter)
            + t.bold
            + "  ".join([f.format(v) for f, v in zip(fmts, fields)])
            + t.reset
        )

        def datestr(t):
            p = pendulum.from_timestamp(t)
            return p.to_datetime_string()

        # table body
        for r in resources:
            values = list(getattr(r, k) for k in fields[:-2])
            values.append(datestr(r.modified))
            values.append(datestr(r.created))
            if not values[1] and r.desc:
                values[1] = r.desc[: widths[1]]
            else:
                values[1] = values[1][: widths[1]]
            if uuid_pfx < 32:
                values[0] = values[0][:uuid_pfx]
            print(
                left_gutter
                + "  ".join([c + f.format(v) for c, f, v in zip(colors, fmts, values)])
                + t.reset
            )
            if relations and len(r.relations) > 0:
                relitems = []
                for rel in r.relations:
                    if rel.subject == r.uuid:
                        fmt = f"{t.white}{{p}}->{t.blue}{{o}}"
                    else:
                        fmt = f"{t.blue}{{s}}->{t.white}{{p}}"
                    item = fmt.format(
                        s=rel.subject[:uuid_pfx],
                        p=rel.predicate,
                        o=rel.object[:uuid_pfx],
                    )
                    relitems.append(item)
                print(f"+-- {' / '.join(relitems)}")
    else:
        items = []
        for r in d.find():
            name_color = "w"
            if r.name:
                name = r.name
            elif r.desc:
                name = r.desc[:40]
                name_color = t.blue
            else:
                name = r.uuid
                name_color = t.green
            item = f"{name_color}{name}{t.yellow}:{r.type}"
            items.append(item)
        if items:
            columnized = _display_in_columns(items, max_line=t.width)
            print(columnized + t.reset)


def _uuid_prefix(uuids, step=4, maxlen=32):
    """Get smallest multiple of `step` len prefix that gives unique values.
    """
    full = set(uuids)
    for n in range(step, maxlen, step):
        prefixes = {u[:n] for u in uuids}
        if len(prefixes) == len(full):
            return n
    return maxlen


def cat_resources(path, objects=(), color=True):
    d = DMF(path=path)
    t = ColorTerm(enabled=color)
    unmatched = set(objects)
    first = True
    # get all resources,
    # display any that match an object as a prefix
    for r in d.find():
        for oid in unmatched:
            if r.uuid.startswith(oid):
                unmatched.remove(oid)  # don't show twice
                if not first:
                    _cat_resource_sep(t)
                _cat_resource_show(t, r)
                first = False
                break


def _cat_resource_sep(t):
    print(f"{t.blue}{'-' * 60}")


def _cat_resource_show(cp, r):
    d = r.as_dict()
    json.dump(d, cp, indent=2)
    print()


# regular expression to find VT100 color escape codes
_noprint_re = re.compile(r"\033\[[0-9]+m")


def _display_in_columns(items, max_line=80, col_sep="  ", row_sep="\n"):
    """Take a list of items and max line width, and calculate  display
    of the items in columns.

    The algorithm is simple, just trying increasing numbers of columns and
    picking the largest number that did not result in a row that was too wide.

    The input items are not re-ordered.

    Args:
        items (List[str]): String items
        max_line (int): Maximum width for any displayed line (row)
        col_sep (str): Separator between columns, after each item
        row_sep (str): Separator between rows, at the end of each line

    Returns:
        str:
    """
    if not items:
        return ""
    # Calculate item lengths, stripping terminal escapes
    lengths, nplengths = [], []
    for item in items:
        clean = _noprint_re.sub("", item)
        lengths.append(len(clean))
        nplengths.append(len(item) - len(clean))
    col_sep_len = len(col_sep)  # useful later
    # Give up immediately, putting everything in one column,
    # if any single item doesn't fit
    if max_line <= max(lengths) + col_sep_len:
        return row_sep.join(items)
    # Determine maximum number of columns
    max_columns = 1  # number of columns
    max_widths = [max_line]  # width of each column
    n = len(lengths)
    # Determine number of columns.
    # Start at 2 columns, stop when cannot fit items side-by-side
    for i in range(2, n):
        # initialize calculated widths of each column
        widths = [0] * i
        # for display where all columns are same length except last,
        # number of items per column is ceiling of items/#col
        nitems = int(math.ceil(n / i))
        # put items in each column
        for col in range(i):
            pad = 0 if col == (i - 1) else col_sep_len  # sep. between columns
            # put items in the current column, adjusting the column
            # max width to widest item
            maxj = min(n, (col + 1) * nitems)  # don't overshoot on last col
            for j in range(col * nitems, maxj):
                widths[col] = max(widths[col], lengths[j] + pad)
        # total width is sum of column widths
        line_len = sum(widths)
        # if we went over, then stop
        if line_len > max_line:
            break
        # otherwise, this is valid -- save and continue
        max_columns, max_widths = i, widths[:]
    # Put items into rows of max. number of columns determined above
    nrows, rows = int(math.ceil(n / max_columns)), []
    for row in range(nrows):
        col, row_items = 0, []
        # skip through items by nrows at a time, to move across the columns,
        # and fill in the items for the current row (which acts as an offset)
        for i in range(row, len(items), nrows):
            # create format string with width = col. max including esc chars,
            # but without padding since we will add that when we join
            # the row items together
            pad = 0 if col == (max_columns - 1) else col_sep_len
            fmt = "{{:{n}s}}".format(n=max_widths[col] + nplengths[i] - pad)
            # format row item for column
            row_items.append(fmt.format(items[i]))
            col += 1  # move to next column
        # append the row items as big string
        rows.append(col_sep.join(row_items))
    # Final result is a big string of the rows joined together
    return row_sep.join(rows)


def _import_resource(filename):
    """Import a resource from 'filename'. Raises a ValueError if that
    fails. Most of the code is simply generating error messages.
    """
    if not os.path.exists(filename):
        raise ValueError('File "{}" not found'.format(filename))
    try:
        f = open(filename)
    except Exception as e:
        raise ValueError('Cannot open file "{}": {}'.format(filename, e))
    try:
        j = json.load(f)
    except json.JSONDecodeError as e:
        raise ValueError('Cannot parse JSON file "{}": {}'.format(filename, e))
    try:
        r = resource.Resource(value=j)
        r.validate()
    except (ValueError, jsonschema.ValidationError) as err:
        raise ValueError("Invalid resource: {}".format(err))
    return r


def _import_jupyternb(path):
    """Create & import a resource from a Jupyter Notebook file at `path`.
    Assume that `path` exists and is a Jupyter Notebook.

    Args:
        path (str): Jupyter Notebook file.
    Returns:
          (Resource) DMF Resource representing the notebook.
    """
    r = resource.Resource(type_=resource.TY_NOTEBOOK)
    filename = os.path.splitext(os.path.split(path)[1])[0]
    # XXX: add notebook 'metadata' as FilePath metadata attr
    r.v["datafiles"].append({"desc": filename, "path": path})
    r.v["desc"] = filename
    r.validate()
    return r


def _import_python(path):
    """Create & import a resource from a Python file at `path`.
    Assume that `path` exists and is a valid Python file.

    Args:
        path (str): Python file name.
    Returns:
          (Resource) DMF Resource representing the notebook.
    """
    r = resource.Resource(type_=resource.TY_CODE)
    filename = os.path.splitext(os.path.split(path)[1])[0]
    r.v["codes"].append({"name": filename, "language": "python", "type": "module"})
    r.v["datafiles"].append({"desc": filename, "path": path})
    r.validate()
    return r


def _import_file(path):
    """Create & import a resource from a generic file at `path`.
    Assume that `path` exists.
 
    Args:
        path (str): File name.
    Returns:
          (Resource) DMF Resource representing the notebook.
    """
    r = resource.Resource(type_=resource.TY_DATA)
    filename = os.path.split(path)[1]
    r.v["datafiles"].append({"desc": filename, "path": path})
    r.v["desc"] = filename
    r.validate()
    return r



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
# This file contians subroutines related to finding the bounds for big M constraints
# additionally, depedent reaction stoichiometries are detected here
from idaes.surrogate import ripe
import numpy as np


def stoich_cons(stoichs):
    # This subroutine identifies dependent stoichiometries
    smat = np.array(stoichs)
    nt, ns = np.shape(stoichs)
    scons_small = []
    scons_large = []
    fam_cons = []
    nfcons = []
    ncons = []
    for con in [3, 4]:
        for ins in nt.permutations(range(nt), con):  # was it
            inds = list(ins)
            tmat = smat[inds, :]
            rank = np.linalg.matrix_rank(tmat)
            if rank < con:
                msize = count_neg(tmat)
                if all(msize[0] == item for item in msize):
                    fam_cons.append([inds[i] + 1 for i in inds])
                    nfcons.append(con - 1)
                else:
                    bigones = np.argwhere(msize == np.max(msize)).flatten().tolist()
                    smallones = np.argwhere(msize != np.max(msize)).flatten().tolist()
                    scons_small.append([inds[i] + 1 for i in smallones])
                    scons_large.append([inds[i] + 1 for i in bigones])
                    ncons.append(con - 1)
    return scons_small, scons_large, ncons, fam_cons, nfcons


def count_neg(mat):
    r = []
    for i in range(len(mat)):
        vec = mat[i]
        count = 0
        for item in vec:
            # This line is commented out to enforce total reaction order (not just reactants)
            #            if item < 0:
            count += abs(item)
        r.append(count)
    return r


def get_bounds(inargs):
    sharedata = inargs[-1]
    inargs[-1]["maxmiptime"] = 60.0
    pc = inargs[-2]
    # try except commented out
    #    try:
    res = ripe.genpyomo.ripeomo(*inargs)
    #    except:
    #        res = {}
    #        res['maxk'] = sharedata['bounds']['k']['max']
    #        res['maxe'] = sharedata['bounds']['e']['max']
    # Calculate upperbound of E from closed form expression
    # c_1,2 = (-1/(R*T_min,max))
    # E_max = ln(c2/c1)/(c1-c2)
    T = [pc["T"][i][0] for i in range(pc["npc"])]
    Tmax = np.max(T)
    Tmin = np.min(T)
    c1 = -1 / (sharedata["gasconst"] * Tmin)
    c2 = -1 / (sharedata["gasconst"] * Tmax)
    arrpen = c2 - c1
    # sharedata['bounds']['e']['max'] = np.log(c2/c1)/(c1-c2)
    kmax = res["maxk"]
    # emax = res["maxe"]
    if kmax >= sharedata["bounds"]["k"]["max"] or kmax == 0.0:
        kmax = 10 * sharedata["bounds"]["k"]["max"]
    elif kmax == 0.0:
        kmax = 10 * sharedata["bounds"]["k"]["max"]
    else:
        kmax = 10 * kmax
    #    if emax >= sharedata['bounds']['e']['max'] or emax == 0.0:
    #        emax = 5*sharedata['bounds']['e']['max']
    #    else:
    #        emax = 1000*emax
    sharedata["bounds"]["k"]["max"] = kmax
    sharedata["arrpen"] = arrpen
    return sharedata



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################


def confinv(aterm, targets, results, ccon, ptype, pc, sharedata, sigma):  
    # This subroutine calculates 95% confidence intervals for estimated ripe models
    # inputs:
    # aterm     - activity array
    # targets   - observed rates of generation
    # results   - selected ripe model
    # ccon      - selected cardinality
    # ptype     - flag for isothermal/arrhenious
    # pc        - process condition array
    # sharedata - share data array defined in shared.py
    # sigma     - estimated/defined residual variance
    # Outputs:
    # kres      - preexponential factor estimate/confidence
    # eres      - activation eenrgy " "
    # stoichres - selected stoichiometry
    # mechres   - selected mechanisms
    # r2        - R2 on training data

    # Alpha value defined 95% CI, can be made variable with ease
    alpha = 0.05
    import numpy as np
    import ripe
    # from scipy.optimize import curve_fit
    from scipy.stats.distributions import t

    n, ns, nm, nh = np.shape(aterm)
    # Reshape targets and aterm
    flat_targets = np.ndarray.flatten(targets, order="F")
    flat_sigma = np.ndarray.flatten(sigma, order="F")
    # Tdata needs to be appended to aterm in the case of arrhenious
    tempaterm = np.zeros([n, ns, ccon])
    idx = 0
    for res in results["ind"]:
        tempaterm[:, :, idx] = aterm[:, :, int(res[0]) - 1, int(res[1]) - 1]
        idx += 1
    aterm = tempaterm

    # collect results
    p0 = []
    kres = []
    eres = []
    stoichres = []
    mechres = []
    # kinformres = []
    for inds in results["ind"]:
        i1 = inds[0]
        i2 = inds[1]
        mechres.append(i1)
        stoichres.append(i2)
        for k in results["k"]:
            if k[1][0] == i1 and k[1][1] == i2:
                # unwind scaling from model selection here
                #                tempk = s_target * (k[0] / scales['nfactor'][int(i1)-1][int(i2)-1])
                tempk = k[0]
                kres.append(tempk)
                p0.append(tempk)
    if "E" in results.keys():
        for inds in results["ind"]:
            i1 = inds[0]
            i2 = inds[1]
            for e in results["E"]:
                if e[1][0] == i1 and e[1][1] == i2:
                    eres.append(e[0])
                    p0.append(e[0])

    # analyze what kinetic model is used to determine models to include in conf_inv calculation
    st = np.average(flat_targets)
    nf = len(flat_targets)
    p = len(p0)
    dof = max(1, int(nf / ns) - p)
    tval = t.ppf(1.0 - alpha / 2.0, dof)

    # The file kinforms.py contains definitions for types kinetic models
    # defaults include: isothermal, arrhenious, arrhenious-reformulation
    if ptype == "arr":
        if "Tref" not in pc.keys():
            sr = ripe.kinforms.arr(
                targets, np.ndarray.flatten(pc["T"]), aterm, p0, sharedata["gasconst"]
            )  # ,sigma)
            sensmat = ripe.kinforms.arrjac(
                targets, np.ndarray.flatten(pc["T"]), aterm, p0, sharedata["gasconst"]
            )
        if "Tref" in pc.keys():
            sr = ripe.kinforms.refarr(
                targets,
                np.ndarray.flatten(pc["T"]),
                sharedata["Tref"],
                aterm,
                p0,
                sharedata["gasconst"],
            )  # ,sigma)
            sensmat = ripe.kinforms.refarrjac(
                targets,
                np.ndarray.flatten(pc["T"]),
                sharedata["Tref"],
                aterm,
                p0,
                sharedata["gasconst"],
            )
    else:
        # Calculate squared residuals and sensitivity matrix
        sr = ripe.kinforms.lin(targets, aterm, p0)
        sensmat = ripe.kinforms.linjac(targets, aterm, p0)

    sst = sum((flat_targets - st) ** 2)  # ,flat_sigma))
    r2 = 1.0 - sum(sr) / sst

    # Calculate confidence intervals
    stmat = np.matmul(np.ndarray.transpose(sensmat), sensmat)
    mmat = np.matmul(
        np.matmul(np.ndarray.transpose(sensmat), np.diag(flat_sigma)), sensmat
    )
    invmat = np.linalg.inv(stmat)
    covar = np.multiply(np.multiply(invmat, mmat), invmat)
    confinvs = [tval * (s ** 0.5) for s in np.diag(covar)]
    itemp = 0
    if ptype == "arr":
        for i in range(int(0.5 * len(p0))):
            kres[itemp] = [kres[itemp], confinvs[i]]
            eres[itemp] = [eres[itemp], confinvs[int(0.5 * len(p0)) + i]]
            itemp += 1
    else:
        for i in range(len(p0)):
            kres[i] = [kres[i], confinvs[i]]

    return [kres, eres, stoichres, mechres, r2]



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
The `idaes.vis` subpackage contains the framework and implementation
of plots that are expected to be of general utility within the IDAES
framework.

For users, an entry point is provided for IDAES classes to produce
plots with the :class:`idaes.vis.plotbase.PlotRegistry` singleton.

Plots will inherit from the interface in :class:`idaes.vis.plotbase.PlotBase`,
which provides some basic methods.

The current implementations all use the Python "bokeh" package, and can
be found in :mod:`idaes.vis.bokeh_plots`.

For more details, please refer to the visualization section of the
main IDAES documentation.
"""
__all__ = ["bokeh_plots", "plotbase", "plotutils"]



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""Commandline Utilities for Managing the IDAES Data Directory"""

__author__ = "John Eslick"

import click
import idaes.solvers
from idaes.commands import cb

@cb.command(name="get-extensions", help="Get solvers and libraries")
@click.option(
    "--url",
    help="URL to download solver",
    default=idaes._config.default_binary_url)
@click.option("--verbose", help="Show details", is_flag=True)
def get_extensions(url, verbose):
    if url is not None:
        click.echo("Getting files...")
        idaes.solvers.download_binaries(url, verbose)
        click.echo("Done")
    else:
        click.echo("\n* You must provide a download URL for IDAES binary files.")



from .cstr import CSTR
from .flash import Flash
from .gibbs_reactor import GibbsReactor
from .heat_exchanger import HeatExchanger, HeatExchangerFlowPattern
from .heater import Heater
from .heat_exchanger_1D import HeatExchanger1D
from .mixer import Mixer, MomentumMixingType, MixingType
from .plug_flow_reactor import PFR
from .pressure_changer import PressureChanger
from .separator import Separator, SplittingType, EnergySplittingType
from .stoichiometric_reactor import StoichiometricReactor
from .equilibrium_reactor import EquilibriumReactor
from .feed import Feed
from .product import Product
from .feed_flash import FeedFlash
from .statejunction import StateJunction
from .translator import Translator



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for translator block.
Authors: Andrew Lee
"""

import pytest

from pyomo.environ import ConcreteModel

from idaes.core import FlowsheetBlock
from idaes.unit_models.translator import Translator

from idaes.property_models.activity_coeff_models.BTX_activity_coeff_VLE \
    import BTXParameterBlock
from idaes.property_models.examples.saponification_thermo import \
    SaponificationParameterBlock

from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              fixed_variables_set,
                                              activated_constraints_set,
                                              number_unused_variables)
from idaes.core.util.testing import (get_default_solver,
                                     PhysicalParameterTestBlock)


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_default_solver()


# -----------------------------------------------------------------------------
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = Translator(default={
            "inlet_property_package": m.fs.properties,
            "outlet_property_package": m.fs.properties})

    # Check unit config arguments
    assert len(m.fs.unit.config) == 8

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.inlet_property_package is m.fs.properties
    assert m.fs.unit.config.outlet_property_package is m.fs.properties
    assert m.fs.unit.config.outlet_state_defined
    assert not m.fs.unit.config.has_phase_equilibrium


# -----------------------------------------------------------------------------
class TestTranslate(object):
    @pytest.fixture(scope="class")
    def trans(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties1 = SaponificationParameterBlock()
        m.fs.properties2 = BTXParameterBlock()

        m.fs.unit = Translator(
                default={"inlet_property_package": m.fs.properties1,
                         "outlet_property_package": m.fs.properties2})

        return m

    @pytest.mark.build
    def test_build(self, trans):
        assert hasattr(trans.fs.unit, "properties_in")
        assert hasattr(trans.fs.unit, "properties_out")

        assert hasattr(trans.fs.unit, "inlet")
        assert len(trans.fs.unit.inlet.vars) == 4
        assert hasattr(trans.fs.unit.inlet, "flow_vol")
        assert hasattr(trans.fs.unit.inlet, "conc_mol_comp")
        assert hasattr(trans.fs.unit.inlet, "temperature")
        assert hasattr(trans.fs.unit.inlet, "pressure")

        assert hasattr(trans.fs.unit, "outlet")
        assert len(trans.fs.unit.outlet.vars) == 4
        assert hasattr(trans.fs.unit.outlet, "flow_mol")
        assert hasattr(trans.fs.unit.outlet, "mole_frac_comp")
        assert hasattr(trans.fs.unit.outlet, "temperature")
        assert hasattr(trans.fs.unit.outlet, "pressure")

        assert number_variables(trans) == 25
        assert number_total_constraints(trans) == 12
        assert number_unused_variables(trans) == 8

    def test_dof(self, trans):
        trans.fs.unit.inlet.flow_vol.fix(1.0e-03)
        trans.fs.unit.inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        trans.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        trans.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        trans.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
        trans.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fix(0.0)

        assert degrees_of_freedom(trans) == 5

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, trans):
        orig_fixed_vars = fixed_variables_set(trans)
        orig_act_consts = activated_constraints_set(trans)

        trans.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(trans) == 5

        fin_fixed_vars = fixed_variables_set(trans)
        fin_act_consts = activated_constraints_set(trans)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    # No solve, as problem has is missing linking constraints

    @pytest.mark.ui
    def test_report(self, trans):
        trans.fs.unit.report()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
try:
    from pyomo.core.expr.sympy_tools import (
        _prod, _sum, _functionMap, _operatorMap, _pyomo_operator_map)
except ImportError:
    from pyomo.core.base.symbolic import (
        _prod, _sum, _functionMap, _operatorMap, _pyomo_operator_map)

from pyomo.environ import ExternalFunction, Var, Expression, value
from pyomo.core.base.constraint import _ConstraintData, Constraint
from pyomo.core.base.expression import _ExpressionData
from pyomo.core.base.block import _BlockData
from pyomo.core.expr.visitor import StreamBasedExpressionVisitor
from pyomo.core.expr.numeric_expr import (ExternalFunctionExpression,
                                          ExpressionBase)
from pyomo.core.expr import current as EXPR, native_types
from pyomo.core.kernel.component_map import ComponentMap

import sympy
from IPython.display import display, Markdown

import logging
import re

_log = logging.getLogger(__name__)

#TODO<jce> Look into things like sum operator and template expressions

def _add_latex_subscripts(x, s):
    if not "_" in x:
        return "{}_{{ {} }}".format(x, s)
    else:
        return re.sub(r"^(.+_)({.+}|.)", "\\1{{ \\2 ,{} }}".format(s), x)

def deduplicate_symbol(x, v, used):
    """
    Check if x is a duplicated LaTeX symbol if so add incrementing Di subscript

    Args:
        x: symbol string
        v: pyomo object
        used: dictionary of pyomo objects with symbols as keys

    Returns:
        Returns a unique symbol.  If x was not in used keys, returns x,
        otherwise adds exponents to make it unique.
    """
    y = x
    k = 1
    while x in used and id(used[x]) != id(v):
        x = _add_latex_subscripts(y, "D{}".format(k))
        k += 1
        if k > 1000:
            # Either the symbol string is not updating or there are lots of dupes
            break
    used[x] = v
    return x

class PyomoSympyBimap(object):
    """
    This is based on the class of the same name in pyomo.core.base.symbolic, but
    it adds mapping latex symbols to the sympy symbols. This will get you pretty
    equations when using sympy's LaTeX writer.
    """
    def __init__(self):
        self.pyomo2sympy = ComponentMap()
        self.parent_symbol = ComponentMap()
        self.sympy2pyomo = {}
        self.sympy2latex = {}
        self.used = {}
        self.i_var = 0
        self.i_expr = 0
        self.i_func = 0
        self.i = 0

    def getPyomoSymbol(self, sympy_object, default=None):
        return self.sympy2pyomo.get(sympy_object, default)

    def getSympySymbol(self, pyomo_object):
        if pyomo_object in self.pyomo2sympy:
            return self.pyomo2sympy[pyomo_object]
        else:
            return self._add_sympy(pyomo_object)

    def sympyVars(self):
        return self.sympy2pyomo.keys()

    def _add_sympy(self, pyomo_object):
        parent_object = pyomo_object.parent_component()
        if isinstance(parent_object, Var):
            sympy_class = sympy.Symbol
            base_name = "x"
            i = self.i_var
            self.i_var += 1
        elif isinstance(parent_object, Expression):
            sympy_class = sympy.Symbol
            base_name = "u"
            i = self.i_expr
            self.i_expr += 1
        elif isinstance(parent_object, ExternalFunction):
            sympy_class = sympy.Function
            base_name = "func"
            i = self.i_func
            self.i_func += 1
        else:
            raise Exception("Should be Var, Exression, or ExternalFunction")

        if parent_object.is_indexed() and parent_object in self.parent_symbol:
            x = self.parent_symbol[parent_object][0]
            latex_symbol = self.parent_symbol[parent_object][1]
        else:
            x = "{}_{}".format(base_name, i)
            base_latex = getattr(parent_object, "latex_symbol", None)
            if base_latex is not None:
                latex_symbol = deduplicate_symbol(base_latex, parent_object,
                                                  self.used)
            else:
                latex_symbol = x
        if parent_object.is_indexed():
            if parent_object not in self.parent_symbol:
                self.parent_symbol[parent_object] = (x, latex_symbol)
            idx = pyomo_object.index()
            idxl = idx
            if isinstance(idx, tuple):
                idxl = ",".join(idx)
                idx = "|".join(idx)
            x = "{}[{}]".format(x, idx)
            latex_symbol = _add_latex_subscripts(latex_symbol, idxl)
        sympy_obj = sympy_class(x)
        self.pyomo2sympy[pyomo_object] = sympy_obj
        self.sympy2pyomo[sympy_obj] = pyomo_object
        self.sympy2latex[sympy_obj] = latex_symbol
        return sympy_obj


class Pyomo2SympyVisitor(StreamBasedExpressionVisitor):
    """
    This is based on the class of the same name in pyomo.core.base.symbolic, but
    it catches ExternalFunctions and does not decend into named expressions.
    """
    def __init__(self, object_map):
        super(Pyomo2SympyVisitor, self).__init__()
        self.object_map = object_map

    def exitNode(self, node, values):
        if isinstance(node, ExternalFunctionExpression):
            # catch ExternalFunction
            _op = self.object_map.getSympySymbol(node._fcn)
        else:
            if node.__class__ is EXPR.UnaryFunctionExpression:
                return _functionMap[node._name](values[0])
            _op = _pyomo_operator_map.get(node.__class__, None)
        if _op is None:
            return node._apply_operation(values)
        else:
            return _op(*tuple(values))

    def beforeChild(self, node, child):
        # Don't replace native or sympy types
        if type(child) in native_types:
            return False, child
        # We will not descend into named expressions...
        if child.is_expression_type():
            if child.is_named_expression_type():
                # To keep expressions from becoming too crazy complicated
                # treat names expressions like variables.
                return False, self.object_map.getSympySymbol(child)
            else:
                return True, None
        # Replace pyomo variables with sympy variables
        if child.is_potentially_variable():
            return False, self.object_map.getSympySymbol(child)
        # Everything else is a constant...
        return False, value(child)


def sympify_expression(expr):
    """
    Converts Pyomo expressions to sympy expressions.
    This is based on the function of the same name in pyomo.core.base.symbolic.
    The difference between this and the Pymomo is that this one checks if the
    expr argument is a named expression and expands it anyway. This allows named
    expressions to only be expanded if they are the top level object.
    """
    #
    # Create the visitor and call it.
    #
    object_map = PyomoSympyBimap()
    visitor = Pyomo2SympyVisitor(object_map)
    is_expr, ans = visitor.beforeChild(None, expr)
    try: # If I explicitly ask for a named expression then descend into it.
        if expr.is_named_expression_type():
            is_expr = True
    except:
        pass
    if not is_expr: # and not expr.is_named_expression_type():
        return object_map, ans

    return object_map, visitor.walk_expression(expr)

def _add_docs(object_map, docs, typ, head):
    """
    Adds documentation for a set of pyomo components to a markdown table

    Args:
        object_map (PyomoSympyBimap): Pyomo, sympy, LaTeX mapping
        docs: string containing a mardown table
        typ: the class of objects to document (Var, Expression, ExternalFunction)
        head: a string to used in the sybol table heading for this class of objects
    Returns:
        A new string markdown table with added doc rows.
    """
    docked = set() # components already documented, mainly for indexed compoents
    whead = True  # write heading before adding first item
    for i, sc in enumerate(object_map.sympyVars()):
        c = object_map.getPyomoSymbol(sc)
        c = c.parent_component() # Document the parent for indexed comps
        if not isinstance(c, typ): continue
        if whead: # add heading if is first entry in this section
            docs += "**{}** | **Doc** | **Path**\n".format(head)
            whead = False
        if id(c) not in docked:
            docked.add(id(c)) # make sure we don't get a line for every index
            try: #just document the parent of indexed vars
                s = object_map.parent_symbol[c][1]
            except KeyError: # non-indexed vars
                s = object_map.sympy2latex[sc]
            docs += "${}$|{}|{}\n".format(s, c.doc, c)
    return docs

def to_latex(expr):
    """Return a sympy expression for the given Pyomo expression

    Args:
        expr (Expression): Pyomo expression
    Returns:
        (dict): keys: sympy_expr, a sympy expression; where, markdown string
            with documentation table; latex_expr, a LaTeX string representation
            of the expression.
    """
    object_map, sympy_expr = sympify_expression(expr)

    # This next bit documents the expression, could use a lot of work, but
    # for now it generates markdown tables that are resonably readable in a
    # jupyter notebook.
    docs = "\nSymbol | Doc | Path\n ---: | --- | ---\n"
    docs = _add_docs(object_map, docs, Var, "Variable")
    docs = _add_docs(object_map, docs, Expression, "Expression")
    docs = _add_docs(object_map, docs, ExternalFunction, "Function")

    #probably should break this up, but this will do for now.
    return {"sympy_expr": sympy_expr,
            "where":docs,
            "latex_expr":sympy.latex(sympy_expr, symbol_names=object_map.sympy2latex)}

def document_constraints(comp, doc=True, descend_into=True):
    """
    Provides nicely formatted constraint documetntation in markdown format,
    assuming the $$latex math$$ and $latex math$ syntax is supported.

    Args:
        comp: A Pyomo component to document in {_ConstraintData, _ExpressionData,
                _BlockData}.
        doc: True adds a documentation table for each constraint or expression.
                Due to the way symbols are semi-automatiaclly generated, the
                exact symbol definitions may be unique to each constraint or
                expression, if unique LaTeX symbols were not provided
                everywhere in a block.
        descend_into: If True, look in subblocks for constraints.

    Returns:
        A string in markdown format with equations in LaTeX form.
    """
    s = None
    if isinstance(comp, _ExpressionData):
        d = to_latex(comp)
        if doc:
            s = "$${}$$\n{}".format(d["latex_expr"], d["where"])
        else:
            s = "$${}$$".format(d["latex_expr"])
    elif isinstance(comp, _ConstraintData):
        d = to_latex(comp.body)
        if doc:
            s = "$${} \le {}\le {}$$\n{}".format(
                comp.lower, d["latex_expr"], comp.upper, d["where"])
        else:
            s = "$${} \le {}\le {}$$".format(
                comp.lower, d["latex_expr"], comp.upper)
    elif isinstance(comp, _BlockData):
        cs = []
        for c in comp.component_data_objects(
            Constraint, descend_into=descend_into):
            if not c.active:
                continue
            cs.append("## {}".format(c))
            cs.append(document_constraints(c, doc))
        s = "\n".join(cs)
    return s

def ipython_document_constraints(comp, doc=True, descend_into=True):
    """
    See document_constraints, this just directly displays the markdown instead
    of returning a string.
    """
    s = document_constraints(comp, doc, descend_into)
    display(Markdown(s))



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
This module contains data utility function tests.
"""
import os
import warnings
import pytest
import numpy as np
import pyomo.environ as pyo

from pyomo.common.fileutils import this_file_dir

import idaes.dmf.model_data as da

_data_dir = os.path.join(this_file_dir(), "data_files")


def test_map_data():
    data1 = os.path.join(_data_dir, "data1.csv")
    data1_meta = os.path.join(_data_dir, "data1_meta.csv")
    m = pyo.ConcreteModel("Data Test Model")
    m.time = pyo.Set(initialize=[1, 2, 3])
    m.pressure = pyo.Var(m.time, doc="pressure (Pa)", initialize=101325)
    m.temperature = pyo.Var(m.time, doc="temperature (K)", initialize=300)
    m.volume = pyo.Var(m.time, doc="volume (m^3)", initialize=10)

    def retag(tag):
        return tag.replace(".junk", "")

    df, df_meta = da.read_data(
        data1, data1_meta, model=m, rename_mapper=retag, unit_system="mks"
    )

    # Check for expected columns in data and meta data
    assert "T" in df
    assert "P" in df
    assert "V" in df
    assert "T" in df_meta
    assert "P" in df_meta
    assert "V" in df_meta

    # Check that the unit strings updated after conversion
    assert df_meta["T"]["units"] == "kelvin"
    # this next unit is Pa
    assert df_meta["P"]["units"] == "kilogram / meter / second ** 2"
    assert df_meta["V"]["units"] == "meter ** 3"

    # Check that the unit conversions are okay
    assert df["T"]["1901-3-3 12:00"] == pytest.approx(300, rel=1e-4)
    assert df["P"]["1901-3-3 12:00"] == pytest.approx(200000, rel=1e-4)
    assert df["V"]["1901-3-3 12:00"] == pytest.approx(5.187286689, rel=1e-4)

    # Check the mapping of the tags to the model (the 1 key is the time indexed
    # from the model, because the reference is for a time-indexed variable)
    assert pyo.value(df_meta["T"]["reference"][1]) == pytest.approx(300, rel=1e-4)
    assert pyo.value(df_meta["P"]["reference"][1]) == pytest.approx(101325, rel=1e-4)
    assert pyo.value(df_meta["V"]["reference"][1]) == pytest.approx(10, rel=1e-4)


def test_map_data_use_ambient_pressure():
    data1 = os.path.join(_data_dir, "data1.csv")
    data1_meta = os.path.join(_data_dir, "data1_meta.csv")
    m = pyo.ConcreteModel("Data Test Model")
    m.time = pyo.Set(initialize=[1, 2, 3])
    m.pressure = pyo.Var(m.time, doc="pressure (Pa)", initialize=101325)
    m.temperature = pyo.Var(m.time, doc="temperature (K)", initialize=300)
    m.volume = pyo.Var(m.time, doc="volume (m^3)", initialize=10)

    def retag(tag):
        return tag.replace(".junk", "")

    df, df_meta = da.read_data(
        data1,
        data1_meta,
        model=m,
        rename_mapper=retag,
        unit_system="mks",
        ambient_pressure="Pamb",
        ambient_pressure_unit="psi",
    )

    # Check that the unit conversions are okay
    assert df["P"]["1901-3-3 12:00"] == pytest.approx(195886, rel=1e-4)


def test_unit_coversion():
    # spot test some unit conversions and features
    # da.unit_convert(x, frm, to=None, system=None, unit_string_map={},
    #                 ignore_units=[], gauge_pressures={}, atm=1.0):

    p_atm = np.array([1, 2, 3])
    p_psi, unit = da.unit_convert(p_atm, "atm", "psi")

    assert p_psi[0] == pytest.approx(14.7, rel=1e-2)
    assert p_psi[1] == pytest.approx(14.7 * 2, rel=1e-2)
    assert p_psi[2] == pytest.approx(14.7 * 3, rel=1e-2)
    assert unit == "pound_force_per_square_inch"

    # ppb is on the list of units to ignore, and not attempt to convert
    p, unit = da.unit_convert(p_atm, "ppb", "psi")
    assert (p[0], pytest.approx(1, rel=1e-2))
    assert unit == "ppb"

    # psig is on the list of gauge pressures.
    p, unit = da.unit_convert(p_psi, "psig", "atm")

    assert p[0] == pytest.approx(2, rel=1e-1)
    assert p[1] == pytest.approx(3, rel=1e-1)
    assert p[2] == pytest.approx(4, rel=1e-1)

    # check the general system of units conversion
    p_pa, unit = da.unit_convert(p_psi, "psi", system="mks")

    assert p_pa[0] == pytest.approx(101325, rel=1e-1)
    assert unit == "kilogram / meter / second ** 2"  # AKA Pa

    # Test for unit conversion of gauge pressue with different atmosperic
    # pressure values
    p, unit = da.unit_convert(
        p_psi, "psig", "atm", ambient_pressure=np.array([1, 1.1, 1.2])
    )

    assert p[0] == pytest.approx(2, rel=1e-1)
    assert p[1] == pytest.approx(3.1, rel=1e-1)
    assert p[2] == pytest.approx(4.2, rel=1e-1)

    # Again but make sure it works with a scalar to
    p, unit = da.unit_convert(p_psi, "psig", "atm", ambient_pressure=1.2)

    assert p[0] == pytest.approx(2.2, rel=1e-1)
    assert p[1] == pytest.approx(3.2, rel=1e-1)
    assert p[2] == pytest.approx(4.2, rel=1e-1)

    # test custom unit string mapping
    p, unit = da.unit_convert(
        p_psi, "MYPRESSURE", "atm", unit_string_map={"MYPRESSURE": "psi"}
    )

    assert p[0] == pytest.approx(1, rel=1e-1)
    assert p[1] == pytest.approx(2, rel=1e-1)
    assert p[2] == pytest.approx(3, rel=1e-1)

    # Test that a unit that doesn't exist remains unchanged
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        p, unit = da.unit_convert(p_psi, "MYPRESSURE", "atm")
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        assert (
            str(w[-1].message) == "In unit conversion, from unit 'MYPRESSURE'"
            " is not defined. No conversion."
        )

    assert p_psi[0] == pytest.approx(14.7, rel=1e-1)
    assert unit == "MYPRESSURE"



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Workspace classes and functions.
"""
# stdlib
import logging
import os
from typing import List
import uuid
import yaml

# third-party
import jsonschema
import six

# local
from .errors import (
    ParseError,
    WorkspaceError,
    WorkspaceNotFoundError,
    WorkspaceConfNotFoundError,
    WorkspaceConfMissingField,
    WorkspaceCannotCreateError,
)
from .util import yaml_load

__author__ = 'Dan Gunter <dkgunter@lbl.gov>'

_log = logging.getLogger(__name__)

CONFIG_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "id": "http://idaes.org/dmf/workspace-config",
    "type": "object",
    "properties": {
        "_id": {"description": "Unique workspace identifier", "type": "string"},
        "htmldocs": {
            "description": "HTML documentation paths",
            "type": "array",
            "items": {
                "type": "string",
                "description": "directory containing Sphinx HTML docs",
            },
            "default": "{dmf_root}/docs/build/html",
        },
        "description": {
            "description": "A human-readable description of the workspace",
            "type": "string",
        },
        "name": {"description": "A short name for the workspace", "type": "string"},
    },
}


class Fields(object):
    """Workspace configuration fields.
    """

    DOC_HTML_PATH = 'htmldocs'  # path to documentation html dir
    LOG_CONF = 'logging'  # logging config


class Workspace(object):
    """DMF Workspace.

    In essence, a workspace is some information at the root of a directory
    tree, a database (currently file-based, so also in the directory tree)
    of *Resources*, and a set of files associated with these resources.

    **Workspace Configuration**

    When the DMF is initialized, the workspace is given as a path to a
    directory. In that directory is a special file named ``config.yaml``,
    that contains metadata about the workspace. The very existence of a
    file by that name is taken by the DMF code as an indication that the
    containing directory is a DMF workspace::

        /path/to/dmf: Root DMF directory
         |
         +- config.yaml: Configuration file
         +- resourcedb.json: Resource metadata "database" (uses TinyDB)
         +- files: Data files for all resources

    The configuration file is a `YAML`_ formatted file

    .. _YAML: http://www.yaml.org/

    The configuration file defines the following key/value pairs:

        _id
            Unique identifier for the workspace. This is auto-generated by
            the library, of course.
        name
            Short name for the workspace.
        description
            Possibly longer text describing the workspace.
        created
            Date at which the workspace was created, as string in the
            ISO8601 format.
        modified
            Date at which the workspace was last modified, as string in the
            ISO8601 format.
        htmldocs
            Full path to the location of the built (not source) Sphinx HTML
            documentation for the `idaes_dmf` package. See
            DMF Help Configuration for more details.

    There are many different possible "styles" of formatting a list of values
    in YAML, but we prefer the simple block-indented style, where the key is
    on its own line and the values are each indented with a dash:

    .. code-block:: YAML

        _id: fe5372a7e51d498fb377da49704874eb
        created: '2018-07-16 11:10:44'
        description: A bottomless trashcan
        modified: '2018-07-16 11:10:44'
        name: Oscar the Grouch's Home
        htmldocs:
        - '{dmf_root}/doc/build/html/dmf'
        - '{dmf_root}/doc/build/html/models'


    Any paths in the workspace configuration, e.g., for the "htmldocs",
    can use two special variables that will take on values relative to the
    workspace location. This avoids hardcoded paths and makes the workspace
    more portable across environments. ``{ws_root}`` will be replaces with
    the path to the workspace directory, and ``{dmf_root}`` will be replaced
    with the path to the (installed) DMF package.

    The `config.yaml` file will allow keys and values it does not know
    about. These will be accessible, loaded into a Python dictionary,
    via the ``meta`` attribute on the :class:`Workspace` instance.
    This may be useful for
    passing additional user-defined information into the DMF at startup.
    """

    #: Name of configuration file placed in WORKSPACE_DIR
    WORKSPACE_CONFIG = 'config.yaml'
    #: Name of ID field
    ID_FIELD = '_id'

    CONF_NAME = 'name'  #: Configuration field for name
    CONF_DESC = 'description'  #: Configuration field for description
    CONF_CREATED = 'created'  #: Configuration field for created date
    CONF_MODIFIED = 'modified'  #: Configuration field for modified date

    def __init__(self, path, create=False, add_defaults=False, html_paths=None):
        """Load or create a workspace rooted at `path`

        Args:
            (str) path: Path for root of workspace
            (bool) create: Create the workspace if it does not exist.
            (bool) add_defaults: Add default values to new config
            html_paths: One or more paths to HTML docs (or None)
        Raises:
            WorkspaceNotFoundError: if ws is not found (and create is false)
            WorkspaceConfNotFoundError: if ws config is not found (& ~create)
            WorkspaceConfMissingField: if there is no ID field.
            DMFError: Anything else
        """
        path = os.path.abspath(path)
        self._wsdir = path
        self._conf = os.path.join(self._wsdir, self.WORKSPACE_CONFIG)
        self._cached_conf = None
        if create:
            # note: these raise OSError on failure
            try:
                os.mkdir(self._wsdir, 0o770)
            except OSError:
                if not os.path.exists(self._wsdir):
                    raise WorkspaceCannotCreateError(self._wsdir)
                if os.path.exists(self._conf):
                    raise WorkspaceError(
                        'existing configuration would be '
                        'overwritten: {}'.format(self._conf)
                    )
                _log.warning(
                    'Using existing path for new DMF workspace: {}'.format(self._wsdir)
                )
            try:
                self._create_new_config(add_defaults)
            except OSError as err:
                raise WorkspaceError(
                    'while creating new workspace ' 'configuration: {}'.format(err)
                )
        else:
            # assert that the workspace exists
            try:
                assert os.path.isdir(self._wsdir)
            except AssertionError:
                raise WorkspaceNotFoundError(self._wsdir)
            try:
                assert os.path.isfile(self._conf)
            except AssertionError:
                raise WorkspaceConfNotFoundError(self._conf)
        self._install_dir = self._get_install_dir()
        try:
            self._id = self.meta[self.ID_FIELD]
        except KeyError:
            raise WorkspaceConfMissingField(path, self.ID_FIELD, 'ID field')
        if html_paths:
            self.set_doc_paths(html_paths)

    def _create_new_config(self, add_defaults):
        conf = open(self._conf, 'w')  # create the file
        new_id = uuid.uuid4().hex  # create new unique ID
        conf.write('{}: {}\n'.format(self.ID_FIELD, new_id))  # write ID
        conf.close()  # flush and close
        if add_defaults:
            self._configure_defaults()

    def _configure_defaults(self):
        """Add default values to the workspace configuration.

        Note: this will *overwrite* any existing values!
        """
        values = {
            key: val
            for key, (desc, val) in WorkspaceConfig()
            .get_fields(only_defaults=True)
            .items()
        }
        self.set_meta(values)

    @property
    def wsid(self):
        """Get workspace identifier (from config file).

        Returns:
            str: Unique identifier.
        """
        return self._id

    @staticmethod
    def _get_install_dir():
        fpath = os.path.realpath(__file__)
        fdir = os.path.dirname(fpath)
        idir = os.path.join(fdir, '..', '..')
        return os.path.abspath(idir)

    def set_meta(self, values, remove=None):
        # type: (dict, list) -> None
        """Update metadata with new values.
        
        Args:
            values (dict): Values to add or change
            remove (list): Keys of values to remove.
        """
        d = self._read_conf()
        if remove:
            for key in remove:
                if key in d:
                    del d[key]
                else:
                    _log.warning('Cannot remove "{}": no such key'.format(key))
        d.update(values)
        self._write_conf(d)
        self._cached_conf = d

    @property
    def meta(self):
        # type: () -> dict
        """Get metadata.

        This reads and parses the configuration.
        Therefore, one way to force a config refresh is to
        simply refer to this property, e.g.::

             dmf = DMF(path='my-workspace')
             #  ... do stuff that alters the config ...
             dmf.meta  # re-read/parse the config

        Returns:
            (dict) Metadata for this workspace.
        """
        # Re-read configuration from file
        d = self._read_conf()
        # In sections that contain paths,
        # substitute special variables in the values with
        # the paths for this workspace.
        if Fields.DOC_HTML_PATH in d:
            p = d[Fields.DOC_HTML_PATH]
            if isinstance(p, str):
                p = [p]
            d[Fields.DOC_HTML_PATH] = list(map(self._expand_path, p))
        if Fields.LOG_CONF in d:
            # Look through logging configuration tree, and
            # replace paths in any value whose key is 'output'.
            stack = [(k, d) for k in d]
            while stack:
                key, p = stack.pop()
                if key == 'output':
                    p[key] = self._expand_path(p[key])
                elif isinstance(p[key], dict):
                    for key2 in p[key].keys():
                        stack.append((key2, p[key]))
        # Return modified configuration metadata
        return d

    def get_doc_paths(self):
        """Get paths to generated HTML Sphinx docs.

        Returns:
            (list) Paths or empty list if not found.
        """
        paths = self.meta.get(Fields.DOC_HTML_PATH, [])
        if len(paths) and hasattr(paths, 'lower'):
            paths = [paths]  # make a str into a list
        return paths

    def set_doc_paths(self, paths: List[str], replace: bool = False):
        """Set paths to generated HTML Sphinx docs.

        Args:
            paths: New paths to add.
            replace: If True, replace any existing paths. Otherwise merge
                     new paths with existing ones.
        """
        existing_paths = self.get_doc_paths()
        if replace:
            new_paths = list(set(existing_paths + paths))
        else:
            new_paths = paths
        if new_paths != existing_paths:
            self.set_meta({Fields.DOC_HTML_PATH: new_paths})

    def _read_conf(self):
        #: type (None) -> dict
        if self._cached_conf is None:
            _log.debug('Load workspace configuration from "{}"'.format(self._conf))
            conf = open(self._conf, 'r')
            try:
                contents = yaml_load(conf)
            except Exception as err:
                raise ParseError(
                    'Cannot load config file "{f}": {e}'.format(f=self._conf, e=err)
                )
            if contents is None:  # empty file
                contents = {}
            self._cached_conf = contents
        else:
            contents = self._cached_conf
        if isinstance(contents, str):
            raise ParseError('File contents cannot be simple str ({})'.format(contents))
        else:
            return contents.copy()

    def _expand_path(self, path):
        # leave paths bracketed by _underscores_ alone
        if len(path) > 2 and path[0] == '_' and path[-1] == '_':
            return path
        return os.path.realpath(
            path.format(ws_root=self.root, dmf_root=self._install_dir)
        )

    def _write_conf(self, contents):
        #: type (dict) -> None
        conf = open(self._conf, 'w')
        yaml.dump(contents, conf, default_flow_style=False)
        conf.close()

    @property
    def root(self):
        """Root path for this workspace.
        This is the path containing the configuration file.
        """
        return self._wsdir

    @property
    def configuration_file(self):
        """Configuration file path.
        """
        return self._conf

    @property
    def name(self):
        return self.meta.get(self.CONF_NAME, 'none')

    @property
    def description(self):
        return self.meta.get(self.CONF_DESC, 'none')


class WorkspaceConfig(object):
    DEFAULTS = {'string': '', 'number': 0, 'boolean': False, 'array': []}

    def __init__(self):
        self._schema = jsonschema.Draft4Validator(CONFIG_SCHEMA)

    def get_fields(self, only_defaults=False):
        """Get all possible metadata fields for workspace config.

        These values come out of the configuration schema.
        Keys starting with a leading underscore, like '_id', are
        skipped.

        Args:
            only_defaults: Only include fields that have a default
                                value in the schema.
        Returns:
            dict: Keys are field name, values are (field description, value).
                  The 'value' gives a default value. Its type is either
                  a list, a number, bool, or a string; the list may be empty.
        """
        prop = CONFIG_SCHEMA['properties']
        result = {}

        for key, item in six.iteritems(prop):
            if key.startswith('_'):
                continue
            desc = item['description']
            type_ = item['type']
            # morph unknown types to string
            if type_ not in self.DEFAULTS:
                _log.warning(
                    'Unknown schema type "{}".' 'Using "string" instead'.format(type_)
                )
                type_ = 'string'
            # figure out value type, look for default values
            default_value = None
            if type_ == 'array':
                itype = item['items']['type']
                idesc = item['items']['description']
                desc = '{}. Each item is a {}'.format(desc, idesc)
                type_ = itype
                if 'default' in item['items']:
                    default_value = [item['items']['default']]
                elif 'default' in item:
                    # default is comma-separated list
                    default_value = item['default'].split(',')
                else:
                    default_value = []
            elif 'default' in item:
                default_value = item['default']

            if default_value is None and only_defaults:
                continue

            if default_value is None:
                default_value = self.DEFAULTS[type_]
            result[key] = (desc, default_value)

        return result


def find_workspaces(root):
    """Find workspaces at or below 'root'.

    Args:
        root (str): Path to start at

    Returns:
        List[str]: paths, which are all workspace roots.
    """
    w = []
    # try all subdirectories of root
    for (dirpath, dirnames, filenames) in os.walk(root):
        for dmfdir in dirnames:
            dmfpath = os.path.join(dirpath, dmfdir)
            conf_file = os.path.join(dmfpath, Workspace.WORKSPACE_CONFIG)
            if os.path.exists(conf_file):
                w.append(dmfpath)
    # also try root itself
    if os.path.exists(os.path.join(root, Workspace.WORKSPACE_CONFIG)):
        w.insert(0, root)
    return w



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################

"""
This module contains functions to read and manage data for use in parameter
esitmation, data reconciliation, and validation.
"""

__author__ = "John Eslick"

import logging
import csv
import pandas as pd
import numpy as np
import pint

import pyomo.environ as pyo
import warnings


def _strip(tag):
    """
    Tag renaming function to remove whitespace, depending on the csv format
    column heading items can pick up some extra whitespace when read into Pandas
    """
    return tag.strip()


_log = logging.getLogger(__file__)

# Some common unit string conversions, these are ones we've come across that
# are not handled by pint. We can organize and refine known unit strings
# more in the future.
_unit_strings = {
    # Pressure
    "PSI": "psi",
    "PSIA": "psi",
    "psia": "psi",
    "PSIG": "psig",
    "INWC": "in water",
    "IN WC": "in water",
    "IN/WC": "in water",
    '" H2O': "in water",
    "INHG": "in hg",
    "IN HG": "in hg",
    "IN/HG": "in hg",
    "HGA": "in hg",
    "IN HGA": "in hg",
    # Fraction
    "PCT": "percent",
    "pct": "percent",
    "PERCT": "percent",
    "PERCT.": "percent",
    "PCNT": "percent",
    "PPM": "ppm",
    "PPB": "ppb",
    "% OPEN": "percent open",
    "% CLSD": "percent closed",
    "% CLOSED": "percent closed",
    # Length
    "IN": "in",
    "INS": "in",
    "INCHES": "in",
    "Inches": "in",
    "FT": "ft",
    "FEET": "ft",
    "FOOT": "ft",
    "Feet": "ft",
    "MILS": "minch",
    # Speed
    "MPH": "mile/hr",
    "IPS": "in/s",
    # Volume
    "KGAL": "kgal",
    # Vol Flow
    "GPM": "gal/min",
    "gpm": "gal/min",
    "CFM": "ft^3/min",
    "KCFM": "ft^3/mmin",
    "SCFM": "ft^3/min",
    "KSCFM": "ft^3/mmin",  # be careful with this one
    # I don't know how to indicate its
    # a volumetric flow at standard
    # conditions
    # Angle
    "DEG": "deg",
    # Angular Speed
    "RPM": "rpm",
    # Fequency
    "HZ": "hz",
    # Temperature
    "DEG F": "degF",
    "Deg F": "degF",
    "deg F": "degF",
    "DEG C": "degC",
    "Deg C": "degC",
    "deg C": "degC",
    "DEGF": "degF",
    "DegF": "degF",
    "DEGC": "degC",
    "DegC": "degC",
    # Temperature Difference
    "DELTA DEG F": "delta_degF",
    "DETLA Deg F": "delta_degF",
    "DETLA deg F": "delta_degF",
    "DETLA DEG C": "delta_degC",
    "DETLA Deg C": "delta_degC",
    "DELTA deg C": "delta_degC",
    "DELTA DEGF": "delta_degF",
    "DELTA DegF": "delta_degF",
    "DELTA degF": "delta_degF",
    "DELTA DEGC": "delta_degC",
    "DELTA DegC": "delta_degC",
    "DELTA degC": "delta_degC",
    "Delta DEG F": "delta_degF",
    "Delta Deg F": "delta_degF",
    "Delta deg F": "delta_degF",
    "Delta DEG C": "delta_degC",
    "Delta Deg C": "delta_degC",
    "Delta deg C": "delta_degC",
    "Delta DEGF": "delta_degF",
    "Delta DegF": "delta_degF",
    "Delta degF": "delta_degF",
    "Delta DEGC": "delta_degC",
    "Delta DegC": "delta_degC",
    "Delta degC": "delta_degC",
    "delta DEG F": "delta_degF",
    "delta Deg F": "delta_degF",
    "delta deg F": "delta_degF",
    "delta DEG C": "delta_degC",
    "delta Deg C": "delta_degC",
    "delta deg C": "delta_degC",
    "delta DEGF": "delta_degF",
    "delta DegF": "delta_degF",
    "delta degF": "delta_degF",
    "delta DEGC": "delta_degC",
    "delta DegC": "delta_degC",
    "delta degC": "delta_degC",
    # Energy
    "MBTU": "kbtu",
    # Mass
    "MLB": "klb",
    "K LB": "klb",
    "K LBS": "klb",
    "lb.": "lb",
    # Mass flow
    "TPH": "ton/hr",
    "tph": "ton/hr",
    "KLB/HR": "klb/hr",
    "KPPH": "klb/hr",
    # Current
    "AMP": "amp",
    "AMPS": "amp",
    "Amps": "amp",
    "Amp": "amp",
    "AMP AC": "amp",
    # pH
    "PH": "pH",
    # VARS (volt-amp reactive)
    "VARS": "VAR",
    "MVARS": "MVAR",
}

_gauge_pressures = {"psig": "psi", "in water gauge": "in water", "in hg gauge": "in hg"}

_ignore_units = [
    "percent",
    "ppm",
    "ppb",
    "pH",
    "VAR",
    "MVAR",
    "H2O",
    "percent open",
    "percent closed",
]


def unit_convert(
    x,
    frm,
    to=None,
    system=None,
    unit_string_map={},
    ignore_units=[],
    gauge_pressures={},
    ambient_pressure=1.0,
    ambient_pressure_unit="atm",
):
    """Convert the quantity x to a different set of units. X can be a numpy array
    or pandas series. The from unit is translated into a string that pint
    can recognize by first looking in unit_string_map then looking in
    know aliases defined in this file. If it is neither place it will be given
    to pint as-is. This translation of the unit is done so that data can be read
    in with the original provided units.

    Args:
        x (float, numpy.array, pandas.series): quantity to convert
        frm (str): original unit string
        to (str): new unit string, or specify "system"
        system (str): unit system to covert to, or specify "to"
        unit_string_map (dict): keys are unit strings and values are
            corresponding strings that pint can recognize.  This only applies to
            the from string.
        ignore_units (list, or tuple): units to not convert
        gauge_pressures (dict): keys are units strings to be considered gauge
            pressures and the values are corresponding absolute pressure units
        ambient_pressure (float, numpy.array, pandas.series): pressure to add
            to gauge pressure to convert it to absolute pressure. The default
            is 1. The unit is atm by default, but can be changed with the
            ambient_pressure_unit argument.
        ambient_pressure_unit (str): Unit for ambient pressure, default is atm,
            and should be a unit recognized by pint
    Returns:
        (tuple): quantity and unit string
    """
    ureg = pint.UnitRegistry(system=system)
    if frm in unit_string_map:
        frm = unit_string_map[frm]
    elif frm in _unit_strings:
        frm = _unit_strings[frm]
    # Now check for gauge pressure
    gauge = False
    if frm in gauge_pressures:
        gauge = True
        frm = gauge_pressures[frm]
    elif frm in _gauge_pressures:
        gauge = True
        frm = _gauge_pressures[frm]
    q = ureg.Quantity
    if (frm in _ignore_units) or (frm in ignore_units):
        return (x, frm)
    else:
        try:
            ureg.parse_expression(frm)
        except pint.errors.UndefinedUnitError:
            warnings.warn(
                "In unit conversion, from unit '{}' is not defined."
                " No conversion.".format(frm),
                UserWarning,
            )
            return x, frm
    if to is None:
        y = q(x, ureg.parse_expression(frm)).to_base_units()
    else:
        y = q(x, ureg.parse_expression(frm)).to(to)
    if gauge:
        # convert gauge pressure to absolute
        y = y + ambient_pressure * ureg.parse_expression(ambient_pressure_unit)
    return (y.magnitude, str(y.units))


def read_data(
    csv_file,
    csv_file_metadata,
    model=None,
    rename_mapper=None,
    unit_system=None,
    ambient_pressure=1.0,
    ambient_pressure_unit="atm",
):
    """
    Read CSV data into a Pandas DataFrame.

    The data should be in a form where the first row contains column headings
    where each column is labeled with a data tag, and the first column contains
    data point labels or time stamps. The metadata should be in a csv file where
    the first column is the tag name, the second column is the model reference (
    which can be empty), the third column is the tag description, and the fourth
    column is the unit of measure string. Any additional information can be
    added to columns after the fourth column and will be ignored. The units of
    measure should be something that is recognized by pint, or in the aliases
    defined in this file. Any tags not listed in the metadata will be dropped.

    Args:
        csv_file (str): Path of file to read
        csv_file_metadata (str): Path of csv file to read column metadata from
        model (ConcreteModel): Optional model to map tags to
        rename_mapper (function): Optional function to rename tags
        unit_system (str): Optional system of units to atempt convert to
        ambient_pressure (float, numpy.array, pandas.series, str): Optional
            pressure to use to convert gauge pressure to absolute if a string is
            supplied the corresponding data tag is assumed to be ambient pressure
        ambient_pressure_unit (str): Optional ambient pressure unit, should be a
            unit recognized by pint.

    Returns:
        (DataFrame): A Pandas data frame with tags in columns and rows indexed
            by time.
        (dict): Column metadata, units of measure, description, and model
            mapping information.
    """
    # read file
    df = pd.read_csv(csv_file, parse_dates=True, index_col=0)
    # Drop empty columns
    df.drop(df.columns[df.columns.str.contains("Unnamed")], axis=1, inplace=True)
    df.rename(mapper=_strip, axis="columns", inplace=True)
    if rename_mapper:
        # Change tag names in some systematic way with the function rename_mapper
        df.rename(mapper=rename_mapper, axis="columns", inplace=True)
    metadata = {}
    if csv_file_metadata:
        with open(csv_file_metadata, "r") as f:
            reader = csv.reader(f)
            for line in reader:
                tag = line[0].strip()
                if rename_mapper:
                    tag = rename_mapper(tag)
                metadata[tag] = {
                    "reference_string": line[1].strip(),
                    "reference": None,
                    "description": line[2].strip(),
                    "units": line[3].strip(),
                }
    # If a model was provided, map the tags with a reference string to the model
    if model:
        for tag, md in metadata.items():
            if md["reference_string"]:
                try:
                    md["reference"] = pyo.Reference(
                        eval(md["reference_string"], {"m": model})
                    )
                except KeyError:
                    warnings.warn(
                        "Tag reference {} not found".format(md["reference_string"]),
                        UserWarning,
                    )
    # Drop the columns with no metadata (assuming those are columns to ignore)
    for tag in df:
        if tag not in metadata:
            df.drop(tag, axis=1, inplace=True)

    # Check if a data tag was specified to use as ambient pressure in conversion
    # of gauge pressures.  If so, get the numbers and replace the tag string
    if isinstance(ambient_pressure, str):
        try:
            ambient_pressure = np.array(df[ambient_pressure])
        except KeyError:
            _log.exception(
                "Tag '{}' does not exist for ambient pressure".format(ambient_pressure)
            )
            raise

    # If unit_system is specified bulk convert everything to that system of units
    # also update the meta data
    if unit_system:
        for tag in df:
            df[tag], metadata[tag]["units"] = unit_convert(
                df[tag],
                metadata[tag]["units"],
                system=unit_system,
                ambient_pressure=ambient_pressure,
                ambient_pressure_unit=ambient_pressure_unit,
            )

    return df, metadata



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Calculates dimensionless data
"""

import math

molecule = ""
critT, critD, critP, acc, R, M, Rm = [0, 0, 0, 0, 0, 0, 0]


def molData(fluidData, mol, RVal):
    """
    Sets up important global values
    """
    global critT, critP, critD, M, triple, acc, R, Rm, molecule
    (critT, critP, critD, M, triple, acc) = fluidData
    molecule = mol
    R = RVal
    Rm = RVal


def PVT(x):
    """
    Calculate dimensionless compressibility
    Inputs:
        X = [Pressure, Density, Temperature]
    OutputS:
        X = [Delta, Tau, Compressibility]
    """
    Pressure = float(x[0])
    Density = float(x[1])
    Temperature = float(x[2])

    Tau = float(critT) / Temperature
    Delta = Density / float(critD)

    if Density != 0:
        Z = (Pressure * 1000 / R / Temperature / Density) - 1.00
    else:
        Z = 0
    return [Delta, Tau, Z]


def P(x):
    """
    Calculate reduced density and inverse reduced temperature
    Return array of Delta, Tau, Pressure
    """
    Temperature = float(x[2])
    Density = float(x[1])
    Pressure = float(x[0])
    Tau = float(critT) / Temperature
    Delta = Density / float(critD)
    return [Delta, Tau, Pressure]


def CP(x):
    """
    Calculate dimensionless isobaric heat capacity
    Inputs:
        X = [Density, Temperature, Isobaric Heat Capacity]
    Outputs:
        X = [Delta, Tau, CP]
    """
    Delta = float(x[0]) / float(critD)
    Tau = float(critT) / float(x[1])
    CP = float(x[2]) / R

    return [Delta, Tau, CP]


def CV(x):
    """
    Calculate dimensionless isochoric heat capacity
    Inputs:
        X = [Density, Temperature, Isochoric Heat Capacity]
    Outputs:
        X = [Delta, Tau, CV]
    """
    Delta = float(x[0]) / float(critD)
    Tau = float(critT) / float(x[1])
    CV = float(x[2]) / R  
    return [Delta, Tau, CV]


def SND(x):
    """
    Calculate dimensionless speed of sound
    Inputs:
        X = [Density, Temperature, Speed of Sound]
    Outputs:
        X = [Delta, Tau, W]
    """
    Delta = float(x[0]) / float(critD)
    Tau = float(critT) / float(x[1])
    w = (float(x[2]) ** 2) / R / 1000 * M / float(x[1])
    return [Delta, Tau, w]


def CP0(x): 
    """
    Calculate dimensionless ideal isobaric heat capacity
    """
    Cp0 = float(x[1]) / Rm / 1000.0
    Temp = float(x[0]) / 1000.0
    return [Temp, Cp0]


def DL(x):
    """
    Calculate Theta and Delta for saturated liquid density
    Inputs:
         X = [Density, Temperature]
    """
    Theta = 1 - float(x[1]) / float(critT)
    Delta = float(x[0]) / float(critD) - 1
    return [Theta, Delta]


def DV(x):
    """
    Calculate Theta and Delta for saturated vapor density
    Inputs:
         X = [Density, Temperature]
    """
    Theta = 1 - float(x[1]) / float(critT)
    Delta = math.log(float(x[0]) / float(critD))
    return [Theta, Delta]


def Dsat(x):
    """
    Calculate dimensionless terms 
    """
    Theta = float(critT) / float(x[1])
    Delta = float(x[0]) / float(critD)
    return [Delta, Theta]


def PV(x):
    """
    Calculate Theta, Tau, and Psi for saturated liquid density
    Inputs:
         X = [Pressure, Temperature]
    """
    Theta = 1 - float(x[1]) / float(critT)
    Psi = math.log(float(x[0]) / float(critP))
    Tau = float(critT) / float(x[1])
    return [Tau, Theta, Psi]



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Test plots (at least creation, not display) in idaes.vis.bokeh_plots
"""
# third-party
from pandas import DataFrame
# package
from idaes.vis import bokeh_plots
from idaes.vis.plotutils import HENStreamType


exchangers = [
    {'hot': 'H2', 'cold': 'C1', 'Q': 1400, 'A': 159, 'annual_cost': 28358, 'stg': 2},
    {'hot': 'H1', 'cold': 'C1', 'Q': 667, 'A': 50, 'annual_cost': 10979, 'stg': 3},
    {'hot': 'H1', 'cold': 'C1', 'Q': 233, 'A': 10, 'annual_cost': 4180, 'stg': 1},
    {'hot': 'H1', 'cold': 'C2', 'Q': 2400, 'A': 355, 'annual_cost': 35727, 'stg': 2},
    {
        'hot': 'H2',
        'cold': 'W',
        'Q': 400,
        'A': 50,
        'annual_cost': 10979,
        'stg': 3,
        'utility_type': HENStreamType.cold_utility,
    },
    {
        'hot': 'S',
        'cold': 'C2',
        'Q': 450,
        'A': 50,
        'annual_cost': 0,
        'stg': 1,
        'utility_type': HENStreamType.hot_utility,
    },
]

streams = [
    {'name': 'H2', 'temps': [423, 423, 330, 303], 'type': HENStreamType.hot},
    {'name': 'H1', 'temps': [443, 435, 355, 333], 'type': HENStreamType.hot},
    {'name': 'C1', 'temps': [408, 396, 326, 293], 'type': HENStreamType.cold},
    {'name': 'C2', 'temps': [413, 413, 353, 353], 'type': HENStreamType.cold},
]


def test_heat_exchanger():
    bokeh_plots.HeatExchangerNetwork(
        exchangers, streams, mark_temperatures_with_tooltips=True
    )

# profile_data = DataFrame(columns=[range(10),)
# def test_profile():
#     bokeh_plots.ProfilePlot()


##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
# This file applies the ALAMOpython module to the ackley problem
# More information on this problemc an be found at
# https://www.sfu.ca/~ssurjano/ackley.html
# The function is corrupted by normally distributed noise of varaince 0.1


from idaes.surrogate import alamopy
from idaes.examples.alamo_python import examples

#Import additional python modules for creating the synthetic data
import math
import numpy as np

def main():
    # Specify number of poitns to be used in the training set
    # Validation data can be provided optionally
    ndata= 50
    nval = 500
    lb = [-15,-15]
    ub = [15,15]
    x = np.random.uniform(lb,ub,(ndata,2))
    xval = np.random.uniform(lb,ub,(nval,2))
    z = [0]*ndata
    zval = [0]*nval
    # specify simulator as examples.sixcamel
    sim = examples.ackley
    for i in range(ndata):
        z[i]=sim(x[i][0],x[i][1])
    for i in range(nval):
        zval[i] = sim(xval[i][0],xval[i][1])

    # Use alamopy's python function wrapper to avoid using ALAMO's I/O format
    almsim = alamopy.wrapwriter(sim)
    # Call alamo through the alamopy wrapper
    res = alamopy.doalamo(x,z,xval=xval,zval=zval,almname='ackley',monomialpower=(1,2,3,4,5,6),expfcns=1,multi2power=(1,2),expandoutput=True)
    # Calculate confidence intervals
    conf_inv = alamopy.almconfidence(res)

    print('Model: {}'.format(res['model']))
    print('Confidence Intervals : {}'.format(conf_inv['conf_inv']))

if __name__ == '__main__':
    np.random.seed(100)
    main()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Demonstration and test flowsheet for a dynamic flowsheet.

"""

# Import Pyomo libraries
from pyomo.environ import (ConcreteModel, SolverFactory, TransformationFactory,
                           Constraint, maximize, Objective)
from pyomo.network import Arc

# Import IDAES core
from idaes.core import FlowsheetBlock

# Import Unit Model Modules
import idaes.property_models.examples.saponification_thermo as thermo_props
import idaes.property_models.examples.saponification_reactions as \
    reaction_props

# Import Unit Model Modules
from idaes.unit_models import CSTR


def main():
    """
    Make the flowsheet object, fix some variables, and solve the problem
    """
    # Create a Concrete Model as the top level object
    m = ConcreteModel()

    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(default={"dynamic": False})

    # Add property packages to flowsheet library
    m.fs.thermo_params = thermo_props.SaponificationParameterBlock()
    m.fs.reaction_params = reaction_props.SaponificationReactionParameterBlock(
        default={"property_package": m.fs.thermo_params})

    # Create unit models
    m.fs.Tank1 = CSTR(default={"property_package": m.fs.thermo_params,
                               "reaction_package": m.fs.reaction_params,
                               "has_equilibrium_reactions": False,
                               "has_heat_of_reaction": True,
                               "has_heat_transfer": True,
                               "has_pressure_change": False})
    m.fs.Tank2 = CSTR(default={"property_package": m.fs.thermo_params,
                               "reaction_package": m.fs.reaction_params,
                               "has_equilibrium_reactions": False,
                               "has_heat_of_reaction": True,
                               "has_heat_transfer": True,
                               "has_pressure_change": False})

    # Make Streams to connect units
    m.fs.stream = Arc(source=m.fs.Tank1.outlet,
                      destination=m.fs.Tank2.inlet)

    TransformationFactory("network.expand_arcs").apply_to(m)

    # Set inlet and operating conditions, and some initial conditions.
    m.fs.Tank1.inlet.flow_vol[0].fix(1.0)
    m.fs.Tank1.inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
    m.fs.Tank1.inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
    m.fs.Tank1.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
    m.fs.Tank1.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
    m.fs.Tank1.inlet.conc_mol_comp[0, "Ethanol"].fix(0.0)

    m.fs.Tank1.inlet.temperature[0].fix(303.15)
    m.fs.Tank1.inlet.pressure[0].fix(101325.0)

    m.fs.Tank1.volume.fix(1.0)
    m.fs.Tank1.heat_duty.fix(0.0)

    m.fs.Tank2.volume.fix(1.0)
    m.fs.Tank2.heat_duty.fix(0.0)

    # Initialize Units
    m.fs.Tank1.initialize()
    m.fs.Tank2.initialize(state_args={
            "flow_vol": 1.0,
            "conc_mol_comp": {"H2O": 55388.0,
                              "NaOH": 100.0,
                              "EthylAcetate": 100.0,
                              "SodiumAcetate": 0.0,
                              "Ethanol": 0.0},
            "temperature": 303.15,
            "pressure": 101325.0})

    # =========================================================================
    # Adding Optimization
    m.fs.obj = Objective(
            expr=m.fs.Tank2.outlet.conc_mol_comp[0, "SodiumAcetate"],
            sense=maximize)

    # Add additional constraint
    m.fs.volume_constraint = Constraint(
            expr=m.fs.Tank1.volume[0] + m.fs.Tank2.volume[0] == 3.0)

    # Set bounds for variables
    m.fs.Tank1.volume[0].setlb(0.5)
    m.fs.Tank1.volume[0].setub(5.0)
    m.fs.Tank2.volume[0].setlb(0.5)
    m.fs.Tank2.volume[0].setub(5.0)

    # Unfix heat inputs
    m.fs.Tank1.volume.unfix()
    m.fs.Tank2.volume.unfix()
    # =========================================================================

    # Create a solver
    solver = SolverFactory('ipopt')
    results = solver.solve(m, tee=True)

    # Print results
    print(results)
    print()
    print("Results")
    print()
    print("Tank 1 Volume")
    m.fs.Tank1.volume.display()
    print()
    print("Tank 2 Volume")
    m.fs.Tank2.volume.display()
    print()
    print("Tank 1 Outlet")
    m.fs.Tank1.outlet.display()
    print()
    print("Tank 2 Outlet")
    m.fs.Tank2.outlet.display()

    # For testing purposes
    return(m, results)


if __name__ == "__main__":
    main()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for process_base.

Author: Andrew Lee
"""
import pytest

from pyomo.environ import Block, ConcreteModel

from idaes.core.process_base import ProcessBaseBlock
from idaes.core import (FlowsheetBlockData,
                        declare_process_block_class)
from idaes.core.util.exceptions import ConfigurationError


@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(FlowsheetBlockData, self).build()


def test_flowsheet():
    # Test flowsheet method
    m = ConcreteModel()
    m.a = Flowsheet()

    assert m.a.flowsheet() is None

    m.b = Block()
    m.b.c = Flowsheet()
    assert m.b.c.flowsheet() is None

    m.a.d = Flowsheet()
    assert m.a.d.flowsheet() is m.a

    m.a.e = Flowsheet([1, 2])
    assert m.a.e[1].flowsheet() is m.a

    m.a.e[1].f = Flowsheet()
    assert m.a.e[1].f.flowsheet() is m.a.e[1]

    m.a.g = Block()
    m.a.g.h = Flowsheet()
    assert m.a.g.h.flowsheet() is m.a

    m.a.i = Block([1, 2])
    m.a.i[1].j = Flowsheet()
    assert m.a.i[1].j.flowsheet() is m.a


def test_get_performance_contents():
    m = ConcreteModel()
    m.b = ProcessBaseBlock()
    assert m.b._get_performance_contents(time_point=0) is None


def test_get_stream_table_contents():
    m = ConcreteModel()
    m.b = ProcessBaseBlock()
    assert m.b._get_stream_table_contents(time_point=0) is None


def test_report():
    # Test that no exceptions occur
    m = ConcreteModel()
    m.b = ProcessBaseBlock()
    m.b.report(dof=True)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################

from pyomo.environ import value, TerminationCondition, SolverStatus
from pyomo.core.base.var import IndexedVar

def solve_successful(status):
    if (status.solver.termination_condition == TerminationCondition.optimal or
        status.solver.status == SolverStatus.ok):
        return True
    return False

def print_ports_summary(ports):
    # find the list of variables to print
    vars_idxs = []
    varname_set = set()
    for prt in ports:
        for vname in prt.vars:
            if vname in varname_set:
                continue

            varname_set.add(vname)
            var = prt.vars[vname]

            for k in var.keys():
                vars_idxs.append((vname, k, '{}[{}]'.format(vname, k)))
    vname_len = max([len(vidx[2]) for vidx in vars_idxs])
    vname_len = max(vname_len, len('Variable'))

    # print the header
    print('{0:>{1}s}'.format('Variable', vname_len), end="\t")

    for prt in ports:
        print(prt.name, end='\t')
    print()

    print(''.join(['-']*vname_len), end='\t')
    for prt in ports:
        print(''.join(['-']*len(prt.name)), end='\t')
    print()

    for vname, idx, vname_idx in vars_idxs:
        print('{0:>{1}s}'.format(vname_idx, vname_len), end="\t")
        for prt in ports:
            print('{0:{1}f}'.format(value(prt.vars[vname][idx]), len(prt.name)), end='\t')
        print()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for StateJunction unit model.
Authors: Andrew Lee
"""

import pytest

from pyomo.environ import (ConcreteModel,
                           SolverStatus,
                           TerminationCondition,
                           value)

from idaes.core import FlowsheetBlock
from idaes.unit_models.statejunction import StateJunction

from idaes.property_models.activity_coeff_models.BTX_activity_coeff_VLE \
    import BTXParameterBlock
from idaes.property_models import iapws95
from idaes.property_models.examples.saponification_thermo import \
    SaponificationParameterBlock

from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              fixed_variables_set,
                                              activated_constraints_set,
                                              number_unused_variables)
from idaes.core.util.testing import (get_default_solver,
                                     PhysicalParameterTestBlock)


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_default_solver()


# -----------------------------------------------------------------------------
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = StateJunction(default={"property_package": m.fs.properties})

    # Check unit config arguments
    assert len(m.fs.unit.config) == 4

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.property_package is m.fs.properties


# -----------------------------------------------------------------------------
class TestSaponification(object):
    @pytest.fixture(scope="class")
    def sapon(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = SaponificationParameterBlock()

        m.fs.unit = StateJunction(
                default={"property_package": m.fs.properties})

        return m

    @pytest.mark.build
    def test_build(self, sapon):
        assert hasattr(sapon.fs.unit, "properties")

        assert hasattr(sapon.fs.unit, "inlet")
        assert len(sapon.fs.unit.inlet.vars) == 4
        assert hasattr(sapon.fs.unit.inlet, "flow_vol")
        assert hasattr(sapon.fs.unit.inlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.inlet, "temperature")
        assert hasattr(sapon.fs.unit.inlet, "pressure")

        assert hasattr(sapon.fs.unit, "outlet")
        assert len(sapon.fs.unit.outlet.vars) == 4
        assert hasattr(sapon.fs.unit.outlet, "flow_vol")
        assert hasattr(sapon.fs.unit.outlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.outlet, "temperature")
        assert hasattr(sapon.fs.unit.outlet, "pressure")

        assert number_variables(sapon) == 8
        assert number_total_constraints(sapon) == 0
        assert number_unused_variables(sapon) == 8

    def test_dof(self, sapon):
        sapon.fs.unit.inlet.flow_vol.fix(1.0e-03)
        sapon.fs.unit.inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        sapon.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        sapon.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        sapon.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
        sapon.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fix(0.0)

        sapon.fs.unit.inlet.temperature.fix(303.15)
        sapon.fs.unit.inlet.pressure.fix(101325.0)

        assert degrees_of_freedom(sapon) == 0

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, sapon):
        orig_fixed_vars = fixed_variables_set(sapon)
        orig_act_consts = activated_constraints_set(sapon)

        sapon.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(sapon) == 0

        fin_fixed_vars = fixed_variables_set(sapon)
        fin_act_consts = activated_constraints_set(sapon)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    # No solve, as problem has no constraints

    @pytest.mark.ui
    def test_report(self, sapon):
        sapon.fs.unit.report()


# -----------------------------------------------------------------------------
class TestBTX(object):
    @pytest.fixture(scope="class")
    def btx(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = BTXParameterBlock(default={"valid_phase": 'Liq'})

        m.fs.unit = StateJunction(default={
            "property_package": m.fs.properties})

        return m

    @pytest.mark.build
    def test_build(self, btx):
        assert hasattr(btx.fs.unit, "properties")

        assert hasattr(btx.fs.unit, "inlet")
        assert len(btx.fs.unit.inlet.vars) == 4
        assert hasattr(btx.fs.unit.inlet, "flow_mol")
        assert hasattr(btx.fs.unit.inlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.inlet, "temperature")
        assert hasattr(btx.fs.unit.inlet, "pressure")

        assert hasattr(btx.fs.unit, "outlet")
        assert len(btx.fs.unit.outlet.vars) == 4
        assert hasattr(btx.fs.unit.outlet, "flow_mol")
        assert hasattr(btx.fs.unit.outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.outlet, "temperature")
        assert hasattr(btx.fs.unit.outlet, "pressure")

        assert number_variables(btx) == 8
        assert number_total_constraints(btx) == 3
        assert number_unused_variables(btx) == 2

    def test_dof(self, btx):
        btx.fs.unit.inlet.flow_mol[0].fix(5)  # mol/s
        btx.fs.unit.inlet.temperature[0].fix(365)  # K
        btx.fs.unit.inlet.pressure[0].fix(101325)  # Pa
        btx.fs.unit.inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        btx.fs.unit.inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        assert degrees_of_freedom(btx) == 0

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, btx):
        orig_fixed_vars = fixed_variables_set(btx)
        orig_act_consts = activated_constraints_set(btx)

        btx.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(btx) == 0

        fin_fixed_vars = fixed_variables_set(btx)
        fin_act_consts = activated_constraints_set(btx)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, btx):
        results = solver.solve(btx)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, btx):
        assert (pytest.approx(5, abs=1e-3) ==
                value(btx.fs.unit.outlet.flow_mol[0]))
        assert (pytest.approx(365, abs=1e-2) ==
                value(btx.fs.unit.outlet.temperature[0]))
        assert (pytest.approx(101325, abs=1e2) ==
                value(btx.fs.unit.outlet.pressure[0]))

    @pytest.mark.ui
    def test_report(self, btx):
        btx.fs.unit.report()


# -----------------------------------------------------------------------------
@pytest.mark.iapws
@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
class TestIAPWS(object):
    @pytest.fixture(scope="class")
    def iapws(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = iapws95.Iapws95ParameterBlock()

        m.fs.unit = StateJunction(default={
                "property_package": m.fs.properties})

        return m

    @pytest.mark.build
    def test_build(self, iapws):
        assert hasattr(iapws.fs.unit, "properties")

        assert len(iapws.fs.unit.inlet.vars) == 3
        assert hasattr(iapws.fs.unit.inlet, "flow_mol")
        assert hasattr(iapws.fs.unit.inlet, "enth_mol")
        assert hasattr(iapws.fs.unit.inlet, "pressure")

        assert hasattr(iapws.fs.unit, "outlet")
        assert len(iapws.fs.unit.outlet.vars) == 3
        assert hasattr(iapws.fs.unit.outlet, "flow_mol")
        assert hasattr(iapws.fs.unit.outlet, "enth_mol")
        assert hasattr(iapws.fs.unit.outlet, "pressure")

        assert number_variables(iapws) == 3
        assert number_total_constraints(iapws) == 0
        assert number_unused_variables(iapws) == 3

    def test_dof(self, iapws):
        iapws.fs.unit.inlet.flow_mol[0].fix(100)
        iapws.fs.unit.inlet.enth_mol[0].fix(4000)
        iapws.fs.unit.inlet.pressure[0].fix(101325)

        assert degrees_of_freedom(iapws) == 0

    @pytest.mark.initialization
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, iapws):
        orig_fixed_vars = fixed_variables_set(iapws)
        orig_act_consts = activated_constraints_set(iapws)

        iapws.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(iapws) == 0

        fin_fixed_vars = fixed_variables_set(iapws)
        fin_act_consts = activated_constraints_set(iapws)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    # No solve, as problem has no constraints

    @pytest.mark.ui
    def test_report(self, iapws):
        iapws.fs.unit.report()
